"""
xpreprocess.py

Centralized script for preprocessing Sth-Sth-v2 for TPU/GCP pretraining, using a multi-stage, multiprocessing strategy.

Mean to run as a standalone script, *prior* to calling `xpretrain.py` =>> mostly because we want to preprocess the data
once, as a fixed cost.
"""
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List

import hydra
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING

from voltron.conf import DatasetConfig
from voltron.overwatch import OverwatchRich
from voltron.preprocessing import index, jsonify_language, preprocess_language, preprocess_videos, unify_batches
from voltron.util import set_global_seed

# Grab Logger
overwatch = logging.getLogger(__file__)


# Set Defaults (Hydra w/ Structured Configs)
DEFAULTS = ["_self_", {"dataset": "sth-sth-v2"}, {"override hydra/job_logging": "overwatch_rich"}]


@dataclass
class PreprocessingConfig:
    # fmt: off
    defaults: List[Any] = field(default_factory=lambda: DEFAULTS)
    hydra: Dict[str, Any] = field(
        default_factory=lambda: {"run": {"dir": "./runs/preprocessing/${now:%m-%d}/dataset-${dataset.name}"}}
    )

    # Command Line Arguments
    seed: int = 21                                  # Random Seed (for reproducibility)
    dry_run: bool = False                           # Dry Run --> Get a sense of preprocessing/serialization footprint

    # Composable / Structured Arguments
    dataset: DatasetConfig = MISSING                # Dataset(s) for pretraining/preprocessing
    # fmt: on


# Hydra Setup :: Retrieve ConfigStore (Singleton) & Register Components
cs = ConfigStore.instance()
cs.store(group="hydra/job_logging", name="overwatch_rich", node=OverwatchRich)  # Annoying - configure logger for Hydra
cs.store(name="config", node=PreprocessingConfig)


@hydra.main(config_path=None, config_name="config")
def xpreprocess(cfg: PreprocessingConfig) -> None:
    overwatch.info("Preprocessing :: Running Phases for Frame Extraction, Language Compilation, and Batching...")

    # Set Randomness
    set_global_seed(cfg.seed)

    # Phase 1 :: Serialize Frames from Video Clips --> Get `registry` for train and val (index structure)
    train_registry, val_registry, train_dir, val_dir = preprocess_videos(
        cfg.dataset.name,
        path=cfg.dataset.path,
        artifact_path=cfg.dataset.artifact_path,
        resolution=cfg.dataset.resolution,
        n_val_videos=cfg.dataset.n_val_videos,
        dry_run=cfg.dry_run,
    )

    # Phase 2 :: Normalize & Tokenize Language  --> Create `index.pt` & `index.json` files
    preprocess_language(
        cfg.dataset.name,
        train_registry,
        val_registry,
        max_lang_len=cfg.dataset.max_lang_len,
        language_model=cfg.dataset.language_model,
        hf_cache=cfg.dataset.hf_cache,
    )
    jsonify_language(train_registry, val_registry)
    index_dir = index(train_registry, val_registry, cfg.dataset.name, artifact_path=cfg.dataset.artifact_path)

    # Phase 3 :: Assemble & Unify Batch "Sets" across the Varied Dataset Formats (for each Model =>> "data-locked")
    unify_batches(
        cfg.dataset.artifact_path,
        cfg.dataset.name,
        train_registry,
        val_registry,
        train_dir,
        val_dir,
        index_dir,
        cfg.dataset.batch_formats,
        max_epochs=cfg.dataset.max_epochs,
        initial_final_alpha=cfg.dataset.initial_final_alpha,
    )


if __name__ == "__main__":
    xpreprocess()



"""
accelerator.py

Base Hydra Structured Configs for defining various accelerator schemes. Uses a simple single inheritance structure.
"""
from dataclasses import dataclass

from hydra.core.config_store import ConfigStore
from omegaconf import MISSING


@dataclass
class AcceleratorConfig:
    accelerator: str = MISSING
    num_accelerators: int = MISSING
    num_workers: int = MISSING


@dataclass
class TPUv2OneConfig(AcceleratorConfig):
    accelerator = "tpu"
    num_accelerators = 1
    num_workers = 4


@dataclass
class TPUv2EightConfig(AcceleratorConfig):
    accelerator = "tpu"
    num_accelerators = 8
    num_workers = 4


@dataclass
class TPUv3OneConfig(AcceleratorConfig):
    accelerator = "tpu"
    num_accelerators = 1
    num_workers = 8


@dataclass
class TPUv3EightConfig(AcceleratorConfig):
    accelerator = "tpu"
    num_accelerators = 8
    num_workers = 8


# Create a configuration group `accelerator` and populate with the above...
cs = ConfigStore.instance()
cs.store(group="accelerator", name="tpu-v2-1", node=TPUv2OneConfig)
cs.store(group="accelerator", name="tpu-v2-8", node=TPUv2EightConfig)
cs.store(group="accelerator", name="tpu-v3-1", node=TPUv3OneConfig)
cs.store(group="accelerator", name="tpu-v3-8", node=TPUv3EightConfig)



"""
checkpointing.py

Core utility class for handling model/optimizer serialization & checkpointing... including (eventually) resume from
checkpoint logic.

Support the following strategies:
    - (k, -1, -1) --> Keep only the most recent "k" epoch checkpoints
    - (k, m, -1) --> Keep the most recent "k" epoch checkpoints and *every* m epoch checkpoint
    - (k, m, s = 2500) --> Keep "k" and "m" subject to above, but also keep *s* step checkpoints for current epoch
"""
import os
from collections import deque
from pathlib import Path
from typing import Any, Optional, Tuple

import wandb


class FixedDeck(deque):
    def __init__(self, maxlen):
        super().__init__(maxlen=maxlen)

    def append(self, x: Any) -> Any:
        pop_value = None
        if self.__len__() == self.maxlen:
            pop_value = self.__getitem__(0)

        # Perform parent append and return popped value, if any!
        super().append(x)
        return pop_value


class CheckpointSaver:
    def __init__(self, strategy: Tuple[int, int, int], run_dir: str, accelerator: str):
        """
        Create a checkpoint saver with the provided strategy that saves to the given path, with handling for the
        specific hardware accelerator.

        :param strategy: Strategy, following the (k, -1, -1) -- (k, m, -1) -- (k, m, s) description above.
        :param run_dir: Path to root of `run_dir`
        :param accelerator: Hardware accelerator to run on -- TODO :: Only TPUs supported right now!
        """
        import torch_xla.core.xla_model as xm

        (self.k, self.m, self.s), self.run_dir, self.accelerator = strategy, run_dir, accelerator
        self.recents, self.intervals, self.step_checkpoints = FixedDeck(maxlen=self.k), set(), set()

        # If `self.s` is -1 --> disable step_checkpoints
        self.enable_step = self.s != -1

        # Create "checkpoints" subdirectory
        self.path = Path(run_dir) / "checkpoints"
        if xm.is_master_ordinal(local=False):
            os.makedirs(self.path, exist_ok=True)

        # Populate `step_checkpoints` on __init__ (if resuming *within* an epoch...)
        self.step_checkpoints.update([c for c in self.path.iterdir() if "local-epoch=" in str(c)])

        # Create Saver
        xm.master_print(f"Created Saver w/ `k` = {self.k}, `m` = {self.m}`, `s` = {self.s}!")

    def save(
        self,
        epoch: int,
        is_local_step: bool,
        model: Any,
        optimizer: Any,
        duration: int,
        local_step: Optional[int] = None,
        train_loss: Optional[float] = None,
        val_loss: Optional[float] = None,
    ) -> None:
        """Performs the save operation, unlinking existing stale checkpoints, if necessary."""
        import torch_xla.core.xla_model as xm

        # Check if saving a `local_step` (within an epoch) or if saving an `epoch`
        if self.enable_step and is_local_step and (local_step % self.s) == 0:
            # Create filename
            step_checkpoint = self.path / f"local-epoch={epoch}-step={local_step}-t={duration}.pt"

            # Perform actual save action...
            #   > IMPORTANT --> XLA/XM will throw an error if optimizer has "param_groups" so only save "state"...
            xm.save([model.state_dict(), optimizer.state_dict()["state"]], step_checkpoint)
            if xm.is_master_ordinal(local=False):
                self.step_checkpoints.add(step_checkpoint)

        elif not is_local_step:
            # Create filename
            if train_loss is None and val_loss is None:
                checkpoint = self.path / f"epoch={epoch}-train=inf-val=inf-t={duration}.pt"
            else:
                checkpoint = self.path / f"epoch={epoch}-train={train_loss:.4f}-val={val_loss:.4f}-t={duration}.pt"

            # Perform actual save action...
            #   > IMPORTANT --> XLA/XM will throw an error if optimizer has "param_groups" so only save "state"...
            xm.save([model.state_dict(), optimizer.state_dict()["state"]], checkpoint)

            if xm.is_master_ordinal(local=False):
                # Conditional Check for M -- Keep if modulated by interval
                if epoch % self.m == 0:
                    self.intervals.add(checkpoint)

                # Remove all "step_checkpoints" now that we successfully made it to the end of the epoch!
                while len(self.step_checkpoints) > 0:
                    os.remove(self.step_checkpoints.pop())

                # Finally, recency add & unlink/delete if necessary
                to_remove = self.recents.append(checkpoint)
                if to_remove is not None and to_remove not in self.intervals:
                    os.remove(to_remove)

                # Extra Redundancy --> Save Model to W&B...
                wandb.save(str(checkpoint))



from .core.vcond import VCond
from .core.vdual import VDual
from .core.vgen import VGen
from .reproductions.vmvp import VMVP
from .reproductions.vr3m import VR3M
from .reproductions.vrn3m import VRN3M



"""
random.py

Utilities for dealing with randomness for PyTorch, across devices (CPU, GPU, TPU).

Loosely inspired by functionality in PyTorch-Lightning:
    > Ref: https://github.com/PyTorchLightning/pytorch-lightning/blob/master/pytorch_lightning/utilities/seed.py

This is pretty important to get right if we're every randomly generating our masks (or prefix dropout) inside our
Dataset __getitem__() with multiple workers... if not handled properly, we will get repeated augmentations anytime
we inject randomness from non-PyTorch sources (e.g., numpy, random)!
    > Ref: https://tanelp.github.io/posts/a-bug-that-plagues-thousands-of-open-source-ml-projects/
"""
import os
import random
from typing import Callable

import numpy as np
import torch

from voltron.util.distributed import get_rank


def set_global_seed(seed: int) -> Callable[[int], None]:
    """Sets seed for all randomness libraries (mostly random, numpy, torch) and produces a `worker_init_fn`"""
    assert np.iinfo(np.uint32).min < seed < np.iinfo(np.uint32).max, "Seed outside the np.uint32 bounds!"

    # Set Seed as an Environment Variable
    os.environ["EXPERIMENT_GLOBAL_SEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

    return worker_init_function


def worker_init_function(worker_id: int) -> None:
    """
    Borrowed directly from PyTorch-Lightning; inspired by this issue comment in the PyTorch repo:
        > Ref: https://github.com/pytorch/pytorch/issues/5059#issuecomment-817392562

    Intuition: You can think of the seed sequence spawn function as a "janky" torch.Generator() or jax.PRNGKey that
    you can run iterative splitting on to get new (predictable) randomness.

    :param worker_id: Identifier for the given worker [0, num_workers) for the Dataloader in question.
    """
    # Get current `rank` (if running distributed) and `process_seed`
    global_rank, process_seed = get_rank(), torch.initial_seed()

    # Back out the "base" (original) seed - the per-worker seed is set in PyTorch:
    #   > https://pytorch.org/docs/stable/data.html#data-loading-randomness
    base_seed = process_seed - worker_id

    # "Magic" code --> basically creates a seed sequence that mixes different "sources" and seeds every library...
    seed_seq = np.random.SeedSequence([base_seed, worker_id, global_rank])

    # Use 128 bits (4 x 32-bit words) to represent seed --> generate_state(k) produces a `k` element array!
    np.random.seed(seed_seq.generate_state(4))

    # Spawn distinct child sequences for PyTorch (reseed) and stdlib random
    torch_seed_seq, random_seed_seq = seed_seq.spawn(2)

    # Torch Manual seed takes 64 bits (so just specify a dtype of uint64
    torch.manual_seed(torch_seed_seq.generate_state(1, dtype=np.uint64)[0])

    # Use 128 Bits for `random`, but express as integer instead of as an array
    random_seed = (random_seed_seq.generate_state(2, dtype=np.uint64).astype(list) * [1 << 64, 1]).sum()
    random.seed(random_seed)



"""
adapt.py

Example script demonstrating how to load a Voltron model (`V-Cond`) and instantiate a Multiheaded Attention Pooling
extractor head for downstream tasks.

This is the basic formula/protocol for using Voltron for arbitrary downstream applications.
"""
import torch
from torchvision.io import read_image

from voltron import instantiate_extractor, load


def adapt() -> None:
    print("[*] Running `adapt` => Demonstrating Voltron for Various Adaptation Applications")

    # Get `torch.device` for loading model (note -- we'll load weights directly onto device!)
    device = "cuda" if torch.cuda.is_available() else "cpu"

    # Load Voltron model --> specify `freeze`, `device` and get model (nn.Module) and preprocessor
    vcond, preprocess = load("v-cond", device=device, freeze=True)

    # Obtain and preprocess an image =>> can be from a dataset, from a camera on a robot, etc.
    img = preprocess(read_image("examples/img/peel-carrot-initial.png"))[None, ...].to(device)
    lang = ["peeling a carrot"]

    # Get various representations...
    with torch.no_grad():
        multimodal_features = vcond(img, lang, mode="multimodal")  # Fused vision & language features
        visual_features = vcond(img, mode="visual")  # Vision-only features (no language)

    # Can instantiate various extractors for downstream applications
    vector_extractor = instantiate_extractor(vcond, n_latents=1)()
    seq_extractor = instantiate_extractor(vcond, n_latents=64)()

    # Assertions...
    assert list(vector_extractor(multimodal_features).shape) == [1, vcond.embed_dim], "Should return a dense vector!"
    assert list(seq_extractor(visual_features).shape) == [1, 64, vcond.embed_dim], "Should return a sequence!"


if __name__ == "__main__":
    adapt()



"""
verify.py

Example script demonstrating how to load all Voltron models (and reproduced models), take input image(s), and get the
various (e.g., multimodal, image-only) representations.

Also serves to verify that representation loading is working as advertised.
"""
import torch
from torchvision.io import read_image

from voltron import load

# Available Models
MODELS = ["v-cond", "v-dual", "v-gen", "r-mvp", "r-r3m-vit", "r-r3m-rn50"]

# Sample Inputs
IMG_A, IMG_B = "examples/img/peel-carrot-initial.png", "examples/img/peel-carrot-final.png"
LANGUAGE = "peeling a carrot"


def verify() -> None:
    print("[*] Running `verify` =>> Verifying Model Representations!")

    # Read both images (we'll use the second image for the dual-frame models)
    image_a, image_b = read_image(IMG_A), read_image(IMG_B)

    # Get `torch.device` for loading model (note -- we'll load weights directly onto device!)
    device = "cuda" if torch.cuda.is_available() else "cpu"

    for model_id in MODELS:
        print(f"\t=> Loading Model ID `{model_id}` and Verifying Representation Shapes!")
        model, preprocess = load(model_id, device=device, freeze=True)

        # Preprocess image, run feature extraction --> assert on shapes!
        if model_id in {"v-cond", "v-cond-base"}:
            for modality, expected in [("multimodal", 196 + 20), ("visual", 196)]:
                representation = model(preprocess(image_a)[None, ...].to(device), [LANGUAGE], mode=modality)
                assert representation.squeeze(dim=0).shape[0] == expected, "Shape not expected!"

        elif model_id in {"v-dual", "v-gen"}:
            for modality, expected in [("multimodal", 196 + 20), ("visual", 196)]:
                dual_img = torch.stack([preprocess(image_a), preprocess(image_b)])[None, ...].to(device)
                representation = model(dual_img, [LANGUAGE], mode=modality)
                assert representation.squeeze(dim=0).shape[0] == expected, "Shape not expected!"

        elif model_id == "r-mvp":
            for mode, expected in [("patch", 196), ("cls", 1)]:
                representation = model(preprocess(image_a)[None, ...].to(device), mode=mode)
                assert representation.squeeze(dim=0).shape[0] == expected, "Shape not expected!"

        elif model_id in {"r-r3m-vit", "r-r3m-rn50"}:
            representation = model(preprocess(image_a)[None, ...].to(device))
            assert representation.squeeze(dim=0).shape[0] == 1, "Shape not expected!"

        else:
            raise ValueError(f"Model {model_id} not supported!")

    # We're good!
    print("[*] All representations & shapes verified! Yay!")


if __name__ == "__main__":
    verify()



"""
tracking.py

Base Hydra Structured Config for defining various run & experiment tracking configurations, e.g., via Weights & Biases.
Uses a simple single inheritance structure.
"""
from dataclasses import dataclass
from typing import List, Optional, Tuple

from hydra.core.config_store import ConfigStore
from omegaconf import MISSING


@dataclass
class TrackingConfig:
    # Generic Logging Frequency --> Matters more for XLA/TPUs... set this to be as large as you can stomach!
    log_frequency: int = 100

    # Checkpointing Strategy --> Save each epoch, keep most recent `idx[0]` checkpoints & *every* `idx[1]` checkpoints
    #   Additionally, save (locally) a checkpoint every `idx[2]` steps for the current epoch (-1).
    checkpoint_strategy: Tuple[int, int, int] = (1, 1, 1500)

    # Weights & Biases Setup
    project: str = "voltron-pretraining"
    entity: str = "voltron-robotics"

    # Notes & Tags are at the discretion of the user... see below
    notes: str = MISSING
    tags: Optional[List[str]] = None

    # Directory to save W&B Metadata & Logs in General -- if None, defaults to `logs/` in the Hydra CWD
    directory: Optional[str] = None


@dataclass
class VoltronTrackingConfig(TrackingConfig):
    # Note: I really like using notes to keep track of things, so will crash unless specified with run.
    #   > For `tags` I like to populate based on other args in the script, so letting it remain None
    notes: str = MISSING


# Create a configuration group `trackers` and populate with the above...
cs = ConfigStore.instance()
cs.store(group="tracking", name="voltron-tracking", node=VoltronTrackingConfig)



"""
optimization.py

General utilities for optimization, e.g., schedulers such as Linear Warmup w/ Cosine Decay for Transformer training.
Notably *does not* use the base PyTorch LR Scheduler, since we call it continuously, across epochs, across steps;
PyTorch has no built-in way of separating the two without coupling to the DataLoader, so may as well make this explicit
in the parent loop.

References
    - MAE: https://github.com/facebookresearch/mae/blob/efb2a8062c206524e35e47d04501ed4f544c0ae8/util/lr_sched.py
    - ⚡️-Bolts: https://github.com/PyTorchLightning/lightning-bolts/blob/master/pl_bolts/optimizers/lr_scheduler.py
"""
import math
from typing import Callable

from torch.optim.optimizer import Optimizer


def get_lr_update(
    opt: Optimizer, schedule: str, lr: float, min_lr: float, warmup_epochs: int, max_epochs: int
) -> Callable[[int, float], float]:
    if schedule == "linear-warmup+cosine-decay":

        def lr_update(epoch: int, fractional_progress: float) -> float:
            """Run the warmup check for linear increase, else cosine decay."""
            if (epoch + fractional_progress) < warmup_epochs:
                new_lr = lr * (epoch + fractional_progress) / max(1.0, warmup_epochs)
            else:
                # Cosine Decay --> as defined in the SGDR Paper...
                progress = ((epoch + fractional_progress) - warmup_epochs) / max(1.0, max_epochs - warmup_epochs)
                new_lr = min_lr + (lr - min_lr) * (0.5 * (1 + math.cos(math.pi * progress)))

            # Apply...
            for group in opt.param_groups:
                if "lr_scale" in group:
                    group["lr"] = new_lr * group["lr_scale"]
                else:
                    group["lr"] = new_lr

            return new_lr

    elif schedule == "none":

        def lr_update(_: int, __: float) -> float:
            return lr

    else:
        raise NotImplementedError(f"Schedule `{schedule}` not implemented!")

    return lr_update



"""
distributed.py

Key distributed utilities; notably provides a standard API for getting relevant data from either CPU/GPU or XLA (TPU)
devices, since the underlying implementation does differ substantially.

Assumes that code is running with one of the following strategies:
    - Single Process (on CPU, GPU)
    - DDP (CPU, GPU)... uses the torch.distributed.launch API & semantics
    - XMP Spawn (TPU)... TPU based XLA + Multiprocessing Spawn semantics

Key Terminology
    -> World Size :: Total number of processes distributed over (# nodes x # devices) -- assumed homogenous!
    -> Rank :: Integer index of current process in the total world size
    -> Local Rank :: Local index on given node in [0, Devices per Node]
"""
from importlib.util import find_spec
from typing import Iterator, TypeVar

import torch
from torch.utils.data import Dataset
from torch.utils.data.distributed import DistributedSampler

T_co = TypeVar("T_co", covariant=True)


def xla_available() -> bool:
    try:
        return find_spec("torch_xla") is not None
    except ModuleNotFoundError:
        return False


def get_rank() -> int:
    """Returns the global rank [0, World Size) of the current process."""
    if xla_available():
        import torch_xla.core.xla_model as xm

        # By default, if XLA is available, assume we're running under XMP Spawn
        return xm.get_ordinal()

    # Try to get rank via torch.distributed, but catch error if only single process
    try:
        return torch.distributed.get_rank()

    # RuntimeError => not running distributed (single process)
    except RuntimeError:
        return 0


class ResumeableDistributedSampler(DistributedSampler):
    def __init__(
        self,
        seen_examples: int,
        resume_epoch: int,
        dataset: Dataset,
        num_replicas: int,
        rank: int,
        shuffle: bool = True,
        seed: int = 0,
    ) -> None:
        super().__init__(dataset, num_replicas=num_replicas, rank=rank, shuffle=shuffle, seed=seed)
        self.seen_examples, self.resume_epoch, self.do_resume = seen_examples, resume_epoch, True

        # Set `seen_examples_per_replica` --> this is necessary for when we re-wrap the iterator in self.__iter__()
        #   > Note: `seen_examples` is across _all_ replicas --> so divide!
        self.seen_examples_per_replica = self.seen_examples // self.num_replicas

    def __iter__(self) -> Iterator[T_co]:
        epoch_iterator = super().__iter__()
        if self.do_resume:
            # Unpack iterator --> list, slice off the first `seen_examples_per_replica` examples, and re-wrap!
            leftover_idxs = list(epoch_iterator)[self.seen_examples_per_replica :]
            return iter(leftover_idxs)
        else:
            return epoch_iterator

    def __len__(self) -> int:
        if self.do_resume:
            # Remove the "seen" sample from self.num_samples; num_samples is *per replica*!
            return self.num_samples - self.seen_examples_per_replica
        else:
            return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        # If epoch != self.resume_epoch --> we're in "regular DistributedSampler" mode (just a wrapper class)
        #   > Intuition: We should *only* truncate examples on the first epoch upon resuming!
        self.epoch = epoch
        if self.epoch != self.resume_epoch:
            self.do_resume = False



"""
models.py

Base Hydra Structured Config for defining various pretraining model architectures and appropriate configurations. Uses a
simple single inheritance structure.
"""
from dataclasses import dataclass
from typing import Tuple

from hydra.core.config_store import ConfigStore
from hydra.utils import to_absolute_path
from omegaconf import MISSING


@dataclass
class ModelConfig:
    arch: str = MISSING
    identifier: str = MISSING

    # Dataset Modality
    data_modality: str = MISSING

    # Default Vision Transformer Configuration
    patch_size: int = 16
    mlp_ratio: float = 4.0

    # Number of examples one can safely fit on an accelerator w/ this model!
    device_bsz: int = MISSING


# @Data-Locked Reproductions --- Encompasses MVP (MAE) + R3M


# MVP (Base Masked Autoencoder)
@dataclass
class MVPConfig(ModelConfig):
    arch: str = "v-mvp"
    identifier: str = MISSING

    # Dataset Modality
    data_modality: str = "state"

    # Base MAE Parameters
    mask_ratio: float = 0.75

    # Architecture Parameters
    encoder_depth: int = MISSING
    encoder_embed_dim: int = MISSING
    encoder_n_heads: int = MISSING

    decoder_depth: int = MISSING
    decoder_embed_dim: int = MISSING
    decoder_n_heads: int = MISSING

    # MAE Loss/Objective Configuration
    norm_pixel_loss: bool = True
    effective_bsz: int = 1024
    device_bsz: int = MISSING

    # Optimization Parameters
    optimizer: str = "adamw"
    schedule: str = "linear-warmup+cosine-decay"
    base_lr: float = 1.5e-4
    min_lr: float = 0.0
    betas: Tuple[float, float] = (0.9, 0.95)
    weight_decay: float = 0.05


@dataclass
class MVPSmallConfig(MVPConfig):
    identifier = "r-mvp"

    # Architecture Parameters -- should match ViT Small Architecture to the letter!
    #   Note: Small is defined in TIMM:
    #       > https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py#L683
    encoder_depth = 12
    encoder_embed_dim = 384
    encoder_n_heads = 6

    decoder_depth = 6
    decoder_embed_dim = 192
    decoder_n_heads = 6

    # Number of examples one can safely fit on an accelerator w/ this model!
    #   > TPU-v3: max of 128 per device.
    device_bsz = 128


# R3M Models --> Just different visual encoders, roughly following the above!
@dataclass
class R3MConfig(ModelConfig):
    arch: str = "v-r3m"
    identifier: str = MISSING

    # Dataset Modality
    data_modality: str = "quintet+language"

    # ViT Architecture Parameters
    depth: int = MISSING
    embed_dim: int = MISSING
    n_heads: int = MISSING

    # Effective Batch Size
    effective_bsz: int = 1024

    # Language Model Parameters
    language_model: str = "distilbert-base-uncased"
    hf_cache: str = to_absolute_path("data/hf-cache")
    language_dim: int = 768
    vocab_size: int = 30522
    reward_dim: int = 1024

    # Loss/Objective Configuration
    lang_reward_weight: float = 1.0
    tcn_weight: float = 1.0
    l1_weight: float = 1e-5
    l2_weight: float = 1e-5
    n_negatives: int = 3
    device_bsz: int = MISSING

    # Optimization Parameters
    optimizer: str = "adam"
    schedule: str = "linear-warmup+cosine-decay"
    lr: float = 1e-4
    min_lr: float = 0.0


@dataclass
class R3MSmallConfig(R3MConfig):
    identifier = "r-r3m-vit"

    # Architecture Parameters -- should match ViT Small Architecture to the letter!
    #   Note: Small is defined in TIMM:
    #       > https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py#L683
    depth = 12
    embed_dim = 384
    n_heads = 6

    # Device Batch Size
    device_bsz = 32


# R3M -- ResNet50 Encoder (instead of ViT)
@dataclass
class ResNet3MConfig(ModelConfig):
    arch: str = "v-rn3m"
    identifier: str = MISSING

    # Dataset Modality
    data_modality: str = "quintet+language"

    # Effective Batch Size
    effective_bsz: int = 1024

    # Architecture Parameters
    fc_dim: int = MISSING

    # Language Model Parameters
    language_model: str = "distilbert-base-uncased"
    hf_cache: str = to_absolute_path("data/hf-cache")
    language_dim: int = 768
    vocab_size: int = 30522
    reward_dim: int = 1024

    # Loss/Objective Configuration
    lang_reward_weight: float = 1.0
    tcn_weight: float = 1.0
    l1_weight: float = 1e-5
    l2_weight: float = 1e-5
    n_negatives: int = 3
    device_bsz: int = MISSING

    # Optimization Parameters
    optimizer: str = "adam"
    lr: float = 1e-4


class RN3M50Config(ResNet3MConfig):
    identifier = "r-r3m-rn50"

    # Architecture Parameters
    fc_dim = 2048

    # Device Batch Size
    device_bsz = 32


# @Voltron Models -- VCond, VDual, VGen


# VCond -- Single Frame + Language Conditioning
@dataclass
class VCondConfig(ModelConfig):
    arch: str = "v-cond"
    identifier: str = MISSING

    # Dataset Modality
    data_modality: str = "state+language"

    # Base MAE Parameters
    mask_ratio: float = 0.75

    # Base Language Parameters --> full sentence dropout only...
    language_model: str = "distilbert-base-uncased"
    hf_cache: str = to_absolute_path("data/hf-cache")
    language_dim: int = 768
    vocab_size: int = 30522
    lang_dropout: float = MISSING

    # Architecture Parameters
    encoder_depth: int = MISSING
    encoder_embed_dim: int = MISSING
    encoder_n_heads: int = MISSING

    decoder_depth: int = MISSING
    decoder_embed_dim: int = MISSING
    decoder_n_heads: int = MISSING

    # MAE Loss/Objective Configuration
    norm_pixel_loss: bool = True
    effective_bsz: int = 1024
    device_bsz: int = MISSING

    # Optimization Parameters
    optimizer: str = "adamw"
    schedule: str = "linear-warmup+cosine-decay"
    base_lr: float = 1.5e-4
    min_lr: float = 0.0
    betas: Tuple[float, float] = (0.9, 0.95)
    weight_decay: float = 0.05


@dataclass
class VCondSmallConfig(VCondConfig):
    identifier = "v-cond"

    # No language dropout...
    lang_dropout = 0.0

    # Architecture Parameters -- should match ViT Small Architecture to the letter!
    #   Note: Small is defined in TIMM:
    #       > https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py#L683
    encoder_depth = 12
    encoder_embed_dim = 384
    encoder_n_heads = 6

    decoder_depth = 6
    decoder_embed_dim = 192
    decoder_n_heads = 6

    # Number of examples one can safely fit on an accelerator w/ this model!
    #   > TPU-v3: max of 128 per device
    device_bsz = 128


@dataclass
class VCondBaseConfig(VCondConfig):
    identifier = "v-cond-base"

    # No language dropout...
    lang_dropout = 0.0

    # Architecture Parameters -- should match ViT Base Architecture to the letter!
    #   Note: Base is defined in TIMM & Original MAE Repository:
    #       > https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py#L723
    #       > https://github.com/facebookresearch/mae/blob/main/models_mae.py#L223
    encoder_depth = 12
    encoder_embed_dim = 768
    encoder_n_heads = 12

    decoder_depth = 8
    decoder_embed_dim = 512
    decoder_n_heads = 16

    # Number of examples one can safely fit on an accelerator w/ this model!
    #   > TPU-v3: max of 128 per device!
    device_bsz = 128


# VDual - Dual Frame (0th Frame + Kth frame) + Language Conditioning
@dataclass
class VDualConfig(ModelConfig):
    arch: str = "v-dual"
    identifier: str = MISSING

    # Dataset Modality
    data_modality: str = "state+ok"

    # Base MAE Parameters
    mae_weight: float = 1.0
    mask_ratio: float = 0.75

    # Base Language Parameters --> full sentence dropout only...
    language_model: str = "distilbert-base-uncased"
    hf_cache: str = to_absolute_path("data/hf-cache")
    language_dim: int = 768
    vocab_size: int = 30522
    lang_dropout: float = MISSING

    # Architecture Parameters
    encoder_depth: int = MISSING
    encoder_embed_dim: int = MISSING
    encoder_n_heads: int = MISSING

    decoder_depth: int = MISSING
    decoder_embed_dim: int = MISSING
    decoder_n_heads: int = MISSING

    # MAE Loss/Objective Configuration -- Cut effective batch size since we see 12-25x contexts per batch example!
    norm_pixel_loss: bool = True
    effective_bsz: int = 1024
    device_bsz: int = MISSING

    # Optimization Parameters
    optimizer: str = "adamw"
    schedule: str = "linear-warmup+cosine-decay"
    base_lr: float = 1.5e-4
    min_lr: float = 0.0
    betas: Tuple[float, float] = (0.9, 0.95)
    weight_decay: float = 0.05


@dataclass
class VDualSmallConfig(VDualConfig):
    identifier = "v-dual"

    # No language dropout...
    lang_dropout = 0.0

    # Architecture Parameters -- should match ViT Small Architecture to the letter!
    #   Note: Small is defined in TIMM:
    #       > https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py#L683
    encoder_depth = 12
    encoder_embed_dim = 384
    encoder_n_heads = 6

    decoder_depth = 6
    decoder_embed_dim = 192
    decoder_n_heads = 6

    # Number of examples one can safely fit on an accelerator w/ this model!
    #   > TPU-v3: max of 128 per device!
    device_bsz = 128


# VGen - Dual Frame with Language Conditioning AND Language Generation
@dataclass
class VGenConfig(ModelConfig):
    arch: str = "v-gen"
    identifier: str = MISSING

    # Dataset Modality
    data_modality: str = "state+ok"

    # Base MAE & LM Parameters --> LM Weight is set such that mae & lang loss are ~same order of magnitude
    mae_weight: float = 1.0
    lm_weight: float = 0.5
    mask_ratio: float = 0.75
    gen_ratio: float = MISSING

    # Base Language Parameters
    language_model: str = "distilbert-base-uncased"
    hf_cache: str = to_absolute_path("data/hf-cache")
    language_dim: int = 768
    vocab_size: int = 30522

    # Architecture Parameters
    encoder_depth: int = MISSING
    encoder_embed_dim: int = MISSING
    encoder_n_heads: int = MISSING

    decoder_depth: int = MISSING
    decoder_embed_dim: int = MISSING
    decoder_n_heads: int = MISSING

    # MAE Loss/Objective Configuration -- Cut effective batch size since we see 12-25x contexts per batch example!
    norm_pixel_loss: bool = True
    effective_bsz: int = 1024
    device_bsz: int = MISSING

    # Optimization Parameters
    optimizer: str = "adamw"
    schedule: str = "linear-warmup+cosine-decay"
    base_lr: float = 1.5e-4
    min_lr: float = 0.0
    betas: Tuple[float, float] = (0.9, 0.95)
    weight_decay: float = 0.05


@dataclass
class VGen50SmallConfig(VGenConfig):
    identifier = "v-gen"

    # LM Parameters --> control % of examples that are for "language generation" (no conditioning)
    gen_ratio = 0.50

    # Architecture Parameters -- should match ViT Small Architecture to the letter!
    #   Note: Small is defined in TIMM:
    #       > https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py#L683
    encoder_depth = 12
    encoder_embed_dim = 384
    encoder_n_heads = 6

    decoder_depth = 6
    decoder_embed_dim = 192
    decoder_n_heads = 6

    # Number of examples one can safely fit on an accelerator w/ this model!
    #   > TPU-v3: max of 64 per device!
    device_bsz = 64


# Create a configuration group `model` and populate with the above...
cs = ConfigStore.instance()

# === @Data-Locked Reproductions ===

# Image-Only MAE/MVP Architectures
cs.store(group="model", name="r-mvp", node=MVPSmallConfig)

# R3M Architectures - ViT & ResNet50
cs.store(group="model", name="r-r3m-vit", node=R3MSmallConfig)
cs.store(group="model", name="r-r3m-rn50", node=RN3M50Config)

# === @Voltron ===

# VCond Architectures
cs.store(group="model", name="v-cond", node=VCondSmallConfig)
cs.store(group="model", name="v-cond-base", node=VCondBaseConfig)

# VDual
cs.store(group="model", name="v-dual", node=VDualSmallConfig)

# VGen
cs.store(group="model", name="v-gen", node=VGen50SmallConfig)



"""
xla_logger.py

Utility class defining various XLA logging methods (called within marked closures), for logging metrics periodically
through training & validation.
"""
from typing import List

import jsonlines
import numpy as np
import torch
import torch_xla.core.xla_model as xm
import wandb


# === Generic (Cross-Model) Epoch End Update ===
def log_epoch_end_update(
    arch: str,
    epoch: int,
    global_step: int,
    run_id: str,
    duration: int,
    train_losses: List[torch.Tensor],
    val_loss: float,
    lr: float,
    step_times: List[float],
) -> None:
    train_loss = torch.stack(list(train_losses)).mean()
    average_step_time = np.mean(list(step_times))

    # Console Logging --> Unclear if it'll work?
    xm.master_print(
        f"Epoch {epoch:03d}, Global Step {global_step:06d} || LR :: {lr:.6f} -- Train Loss :: {train_loss:.4f} "
        f"-- Val Loss :: {val_loss:.4f} -- Total Time (sec) :: {duration}"
    )

    # Get Log-Friendly Arch
    p_arch = {
        "v-mvp": "MVP",
        "v-r3m": "R3M (ViT)",
        "v-rn3m": "R3M (RN)",
        "v-cond": "V-Cond",
        "v-dual": "V-Dual",
        "v-gen": "V-Gen",
    }[arch]

    # Log to Weights & Biases & JSONL
    blob = {
        "Pretrain/Step": global_step,
        "Pretrain/Epoch": epoch,
        "Pretrain/Training Duration": duration,
        "Pretrain/Step Time": average_step_time,
        f"Pretrain/{p_arch} Train Epoch Loss": train_loss.item(),
        f"Pretrain/{p_arch} Train Loss": train_loss.item(),
        f"Pretrain/{p_arch} Validation Loss": val_loss,
        "Pretrain/Learning Rate": lr,
    }

    wandb.log(blob, step=global_step)
    with jsonlines.open(f"{run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
        js_logger.write(blob)


# === Data-Locked Reproductions ===


def log_vmvp_train_update(
    epoch: int,
    global_step: int,
    run_id: str,
    train_losses: List[torch.Tensor],
    lr: float,
    reconstruction_losses: List[torch.Tensor],
    step_times: List[float],
) -> None:
    train_loss = torch.stack(list(train_losses)).mean()
    reconstruction_loss = torch.stack(list(reconstruction_losses)).mean()
    average_step_time = np.mean(list(step_times))

    # Console Logging --> Just log the aggregated train loss...
    xm.master_print(
        f"Epoch {epoch:03d}, Global Step {global_step:06d} || LR :: {lr:.6f} -- Train Loss :: {train_loss:.4f}"
    )

    # Log ot Weights & Biases + JSONL
    blob = {
        "Pretrain/Step": global_step,
        "Pretrain/Epoch": epoch,
        "Pretrain/V-MVP Train Loss": train_loss.item(),
        "Pretrain/Reconstruction Loss": reconstruction_loss.item(),
        "Pretrain/Learning Rate": lr,
        "Pretrain/Step Time": average_step_time,
    }
    wandb.log(blob, step=global_step)
    with jsonlines.open(f"{run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
        js_logger.write(blob)


def log_vr3m_train_update(
    epoch: int,
    global_step: int,
    run_id: str,
    train_losses: List[torch.Tensor],
    lr: float,
    tcn_losses: List[torch.Tensor],
    reward_losses: List[torch.Tensor],
    l1_losses: List[torch.Tensor],
    l2_losses: List[torch.Tensor],
    tcn_accuracies: List[torch.Tensor],
    reward_accuracies: List[torch.Tensor],
    step_times: List[float],
) -> None:
    train_loss = torch.stack(list(train_losses)).mean()
    tcn_loss = torch.stack(list(tcn_losses)).mean()
    reward_loss = torch.stack(list(reward_losses)).mean()
    l1_loss, l2_loss = torch.stack(list(l1_losses)).mean(), torch.stack(list(l2_losses)).mean()
    tcn_accuracy = torch.stack(list(tcn_accuracies)).mean()
    reward_accuracy = torch.stack(list(reward_accuracies)).mean()
    average_step_time = np.mean(list(step_times))

    # Console Logging --> Just log the aggregated train loss...
    xm.master_print(
        f"Epoch {epoch:03d}, Global Step {global_step:06d} || LR :: {lr:.6f} -- Train Loss :: {train_loss:.4f}"
    )

    # Log to Weights & Biases + JSONL
    blob = {
        "Pretrain/Step": global_step,
        "Pretrain/Epoch": epoch,
        "Pretrain/V-R3M Train Loss": train_loss.item(),
        "Pretrain/TCN Loss": tcn_loss.item(),
        "Pretrain/Reward Loss": reward_loss.item(),
        "Pretrain/L1 Loss": l1_loss.item(),
        "Pretrain/L2 Loss": l2_loss.item(),
        "Pretrain/TCN Accuracy": tcn_accuracy.item(),
        "Pretrain/Reward Accuracy": reward_accuracy.item(),
        "Pretrain/Learning Rate": lr,
        "Pretrain/Step Time": average_step_time,
    }
    wandb.log(blob, step=global_step)
    with jsonlines.open(f"{run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
        js_logger.write(blob)


def log_vrn3m_train_update(
    epoch: int,
    global_step: int,
    run_id: str,
    train_losses: List[torch.Tensor],
    lr: float,
    tcn_losses: List[torch.Tensor],
    reward_losses: List[torch.Tensor],
    l1_losses: List[torch.Tensor],
    l2_losses: List[torch.Tensor],
    tcn_accuracies: List[torch.Tensor],
    reward_accuracies: List[torch.Tensor],
    step_times: List[float],
) -> None:
    train_loss = torch.stack(list(train_losses)).mean()
    tcn_loss = torch.stack(list(tcn_losses)).mean()
    reward_loss = torch.stack(list(reward_losses)).mean()
    l1_loss, l2_loss = torch.stack(list(l1_losses)).mean(), torch.stack(list(l2_losses)).mean()
    tcn_accuracy = torch.stack(list(tcn_accuracies)).mean()
    reward_accuracy = torch.stack(list(reward_accuracies)).mean()
    average_step_time = np.mean(list(step_times))

    # Console Logging --> Just log the aggregated train loss...
    xm.master_print(
        f"Epoch {epoch:03d}, Global Step {global_step:06d} || LR :: {lr:.6f} -- Train Loss :: {train_loss:.4f}"
    )

    # Log to Weights & Biases + JSONL
    blob = {
        "Pretrain/Step": global_step,
        "Pretrain/Epoch": epoch,
        "Pretrain/V-RN3M Train Loss": train_loss.item(),
        "Pretrain/TCN Loss": tcn_loss.item(),
        "Pretrain/Reward Loss": reward_loss.item(),
        "Pretrain/L1 Loss": l1_loss.item(),
        "Pretrain/L2 Loss": l2_loss.item(),
        "Pretrain/TCN Accuracy": tcn_accuracy.item(),
        "Pretrain/Reward Accuracy": reward_accuracy.item(),
        "Pretrain/Learning Rate": lr,
        "Pretrain/Step Time": average_step_time,
    }
    wandb.log(blob, step=global_step)
    with jsonlines.open(f"{run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
        js_logger.write(blob)


# === Voltron Models ===
def log_vcond_train_update(
    epoch: int,
    global_step: int,
    run_id: str,
    train_losses: List[torch.Tensor],
    lr: float,
    reconstruction_losses: List[torch.Tensor],
    step_times: List[float],
) -> None:
    train_loss = torch.stack(list(train_losses)).mean()
    reconstruction_loss = torch.stack(list(reconstruction_losses)).mean()
    average_step_time = np.mean(list(step_times))

    # Console Logging --> Just log the aggregated train loss...
    xm.master_print(
        f"Epoch {epoch:03d}, Global Step {global_step:06d} || LR :: {lr:.6f} -- Train Loss :: {train_loss:.4f}"
    )

    # Log to Weights & Biases + JSONL
    blob = {
        "Pretrain/Step": global_step,
        "Pretrain/Epoch": epoch,
        "Pretrain/V-Cond Train Loss": train_loss.item(),
        "Pretrain/Reconstruction Loss": reconstruction_loss.item(),
        "Pretrain/Learning Rate": lr,
        "Pretrain/Step Time": average_step_time,
    }
    wandb.log(blob, step=global_step)
    with jsonlines.open(f"{run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
        js_logger.write(blob)


def log_vdual_train_update(
    epoch: int,
    global_step: int,
    run_id: str,
    train_losses: List[torch.Tensor],
    lr: float,
    reconstruction_losses: List[torch.Tensor],
    zero_reconstruction_losses: List[torch.Tensor],
    k_reconstruction_losses: List[torch.Tensor],
    step_times: List[float],
) -> None:
    train_loss = torch.stack(list(train_losses)).mean()
    reconstruction_loss = torch.stack(list(reconstruction_losses)).mean()
    zero_reconstruction_loss = torch.stack(list(zero_reconstruction_losses)).mean()
    k_reconstruction_loss = torch.stack(list(k_reconstruction_losses)).mean()
    average_step_time = np.mean(list(step_times))

    # Console Logging --> Just log the aggregated train loss...
    xm.master_print(
        f"Epoch {epoch:03d}, Global Step {global_step:06d} || LR :: {lr:.6f} -- Train Loss :: {train_loss:.4f}"
    )

    # Log to Weights & Biases + JSONL
    blob = {
        "Pretrain/Step": global_step,
        "Pretrain/Epoch": epoch,
        "Pretrain/V-Dual Train Loss": train_loss.item(),
        "Pretrain/Reconstruction Loss": reconstruction_loss.item(),
        "Pretrain/Zero Reconstruction Loss": zero_reconstruction_loss.item(),
        "Pretrain/K Reconstruction Loss": k_reconstruction_loss.item(),
        "Pretrain/Learning Rate": lr,
        "Pretrain/Step Time": average_step_time,
    }
    wandb.log(blob, step=global_step)
    with jsonlines.open(f"{run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
        js_logger.write(blob)


def log_vgen_train_update(
    epoch: int,
    global_step: int,
    run_id: str,
    train_losses: List[torch.Tensor],
    lr: float,
    reconstruction_losses: List[torch.Tensor],
    lm_losses: List[torch.Tensor],
    lm_ppl: List[torch.Tensor],
    zero_reconstruction_losses: List[torch.Tensor],
    k_reconstruction_losses: List[torch.Tensor],
    step_times: List[float],
) -> None:
    train_loss = torch.stack(list(train_losses)).mean()
    reconstruction_loss = torch.stack(list(reconstruction_losses)).mean()
    lm_loss = torch.stack(list(lm_losses)).mean()
    lm_perplexity = torch.stack(list(lm_ppl)).mean()
    zero_reconstruction_loss = torch.stack(list(zero_reconstruction_losses)).mean()
    k_reconstruction_loss = torch.stack(list(k_reconstruction_losses)).mean()
    average_step_time = np.mean(list(step_times))

    # Console Logging --> Just log the aggregated train loss...
    xm.master_print(
        f"Epoch {epoch:03d}, Global Step {global_step:06d} || LR :: {lr:.6f} -- Train Loss :: {train_loss:.4f} --"
        f" Reconstruction Loss {reconstruction_loss:.4f} -- LM Loss {lm_loss:.4f}"
    )

    # Log to Weights & Biases + JSONL
    blob = {
        "Pretrain/Step": global_step,
        "Pretrain/Epoch": epoch,
        "Pretrain/V-Gen Train Loss": train_loss.item(),
        "Pretrain/Reconstruction Loss": reconstruction_loss.item(),
        "Pretrain/CLM Loss": lm_loss.item(),
        "Pretrain/CLM Perplexity": lm_perplexity.item(),
        "Pretrain/LM Loss": lm_loss.item(),
        "Pretrain/LM Perplexity": lm_perplexity.item(),
        "Pretrain/Zero Reconstruction Loss": zero_reconstruction_loss.item(),
        "Pretrain/K Reconstruction Loss": k_reconstruction_loss.item(),
        "Pretrain/Learning Rate": lr,
        "Pretrain/Step Time": average_step_time,
    }
    wandb.log(blob, step=global_step)
    with jsonlines.open(f"{run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
        js_logger.write(blob)



"""
vmvp.py

PyTorch Module defining a basic MAE a la Masked Visual Pretraining for Motor Control (MVP), with the requisite
hyperparameters - as defined in the original ImageMAE paper, and as used by both MVP papers.

References:
    - https://github.com/facebookresearch/mae
    - https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py
"""
from typing import Callable, Optional, Tuple

import torch
import torch.nn as nn
from einops import rearrange

from voltron.models.util.optimization import get_lr_update
from voltron.models.util.transformer import Block, PatchEmbed, get_2D_position_embeddings


class VMVP(nn.Module):
    def __init__(
        self,
        resolution: int,
        patch_size: int,
        encoder_depth: int,
        encoder_embed_dim: int,
        encoder_n_heads: int,
        decoder_depth: int,
        decoder_embed_dim: int,
        decoder_n_heads: int,
        optimizer: str,
        schedule: str,
        base_lr: float,
        min_lr: float,
        effective_bsz: float,
        betas: Tuple[float, float],
        weight_decay: float,
        warmup_epochs: int,
        max_epochs: int,
        mask_ratio: float = 0.75,
        mlp_ratio: float = 4.0,
        in_channels: int = 3,
        norm_pixel_loss: bool = True,
    ):
        """
        Initialize an VMVP (MAE) model with the requisite architecture parameters.

        :param resolution: Base image resolution -- usually 224 (ImageNet size).
        :param patch_size: Height/Width of each patch in pixels -- usually 16.
        :param encoder_depth: Number of Transformer blocks in the encoder -- should be greater than decoder.
        :param encoder_embed_dim: Core embedding/hidden dimension for encoder vision transformer backbone.
        :param encoder_n_heads: Number of heads for encoder multi-headed self-attention.
        :param decoder_depth: Number of Transformer blocks in the decoder -- should be relatively shallow.
        :param decoder_embed_dim: Core embedding/hidden dimension for encoder vision transformer backbone.
        :param decoder_n_heads: Number of heads for encoder multi-headed self-attention.
        :param optimizer: String denoting which optimizer to use (for MAEs, usually `adamw`)
        :param schedule: Learning rate schedule to use; for Transformers a linear warmup + decay is recommended!
        :param base_lr: Base learning rate, to be scaled via a linear scaling rule (from scaling laws).
        :param min_lr: Minimum learning rate to decay to over the course of learning (usually 0.0)
        :param effective_bsz: Global batch size for update, dictates the scaling of the base_lr.
        :param betas: Adam optimizer betas (only applicable for `adam` and `adamw`. Prevents early loss spiking.
        :param weight_decay: Weight decay for global weight regularization (only applied to non-bias, non-LN layers).
        :param warmup_epochs: Number of epochs to warmup learning rate for linear warmup schedule.
        :param max_epochs: Total number of training epochs to be run.
        :param mask_ratio: Ratio for number of patches to mask out for MAE -- should be fairly high!
        :param mlp_ratio: Ratio for embedding size to Position-wise FeedForward MLP (gets shrunk back down).
        :param in_channels: Default number of channels in the base image -- almost always 3.
        :param norm_pixel_loss: Normalize decoder pixel targets for reconstruction (better perf, not interpretable).
        """
        super().__init__()
        self.resolution, self.patch_size, self.mask_ratio = resolution, patch_size, mask_ratio
        self.in_channels, self.norm_pixel_loss, self.mlp_ratio = in_channels, norm_pixel_loss, mlp_ratio
        self.optimizer, self.schedule, self.betas, self.weight_decay = optimizer, schedule, betas, weight_decay
        self.lr, self.base_lr, self.min_lr, self.effective_bsz = None, base_lr, min_lr, effective_bsz
        self.warmup_epochs, self.max_epochs = warmup_epochs, max_epochs

        # Encoder/Decoder Parameters
        self.encoder_depth, self.decoder_depth = encoder_depth, decoder_depth
        self.encoder_embed_dim, self.encoder_n_heads = encoder_embed_dim, encoder_n_heads
        self.decoder_embed_dim, self.decoder_n_heads = decoder_embed_dim, decoder_n_heads

        # MAE Encoder Parameters --> MVP uses a CLS Token for feature extraction!
        self.cls_token = nn.Parameter(torch.zeros(1, 1, self.encoder_embed_dim))
        self.patch2embed = PatchEmbed(
            self.resolution, self.patch_size, self.encoder_embed_dim, in_channels=self.in_channels
        )
        self.encoder_pe = nn.Parameter(
            torch.zeros(1, self.patch2embed.num_patches + 1, self.encoder_embed_dim), requires_grad=False
        )
        self.encoder_blocks = nn.ModuleList(
            [Block(self.encoder_embed_dim, self.encoder_n_heads, self.mlp_ratio) for _ in range(self.encoder_depth)]
        )
        self.encoder_norm = nn.LayerNorm(self.encoder_embed_dim, eps=1e-6)

        # Projection from Encoder to Decoder
        self.encoder2decoder = nn.Linear(self.encoder_embed_dim, self.decoder_embed_dim)

        # MAE Decoder Parameters -- Remember the CLS Token!
        self.mask_token = nn.Parameter(torch.zeros(1, 1, self.decoder_embed_dim))
        self.decoder_pe = nn.Parameter(
            torch.zeros(1, self.patch2embed.num_patches + 1, self.decoder_embed_dim), requires_grad=False
        )
        self.decoder_blocks = nn.ModuleList(
            [Block(self.decoder_embed_dim, self.decoder_n_heads, self.mlp_ratio) for _ in range(self.decoder_depth)]
        )
        self.decoder_norm = nn.LayerNorm(self.decoder_embed_dim, eps=1e-6)
        self.decoder_prediction = nn.Linear(self.decoder_embed_dim, (patch_size**2) * in_channels, bias=True)

        # Initialize all Weights
        self.initialize_weights()

    def initialize_weights(self) -> None:
        # Position Encoding -- Fixed 2D Sine-Cosine Embeddings
        enc_pe = get_2D_position_embeddings(self.encoder_embed_dim, int(self.patch2embed.num_patches**0.5), True)
        self.encoder_pe.data.copy_(torch.from_numpy(enc_pe).float().unsqueeze(0))
        dec_pe = get_2D_position_embeddings(self.decoder_embed_dim, int(self.patch2embed.num_patches**0.5), True)
        self.decoder_pe.data.copy_(torch.from_numpy(dec_pe).float().unsqueeze(0))

        # Initialize PatchEmbedding as a Linear...
        nn.init.xavier_uniform_(self.patch2embed.proj.weight.data.view([self.patch2embed.proj.weight.data.shape[0], -1]))

        # Initialize CLS Token & Mask Token w/ Truncated Normal
        nn.init.normal_(self.cls_token, std=0.02)
        nn.init.normal_(self.mask_token, std=0.02)

        # Everything else...
        self.apply(self.transformer_initializer)

    @staticmethod
    def transformer_initializer(m: nn.Module) -> None:
        if isinstance(m, nn.Linear):
            # Use xavier_uniform following Jax ViT
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0.0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.weight, 1.0)
            nn.init.constant_(m.bias, 0.0)

    def mask(
        self, patches: torch.Tensor, mask_ratio: Optional[float] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Perform per-sample random masking by shuffling :: uses argsort random noise to identify masked patches"""
        bsz, n_patches, embed_dim = patches.shape
        if mask_ratio is not None:
            n_keep = int(n_patches * (1 - mask_ratio))
        else:
            n_keep = int(n_patches * (1 - self.mask_ratio))

        # Sample some noise of n_patches size, argsort to get shuffled IDs (keep small), argsort again to get "unshuffle"
        #   > For clarity -- argsort is an invertible transformation (if argsort `restore`, recovers `shuffle`)
        shuffle_idxs = torch.argsort(torch.rand(bsz, n_patches, device=patches.device), dim=1)
        restore_idxs = torch.argsort(shuffle_idxs, dim=1)

        # Get "keep" (visible) patches
        visible_patches = torch.gather(patches, dim=1, index=shuffle_idxs[:, :n_keep, None].repeat(1, 1, embed_dim))

        # Generate the binary mask --> IMPORTANT :: `0` is keep, `1` is remove (following FAIR MAE convention)
        mask = torch.ones(bsz, n_patches, device=patches.device)
        mask[:, :n_keep] = 0
        mask = torch.gather(mask, dim=1, index=restore_idxs)

        return visible_patches, mask, restore_idxs

    def get_representations(self, img: torch.Tensor, mode: str = "patch") -> torch.Tensor:
        """
        Given a single image, extract representations subject to the specified mode in < patch | cls >, where "cls"
        denotes extracting the <CLS> token embedding; for our experiments, we find that running multiheaded attention
        pooling on top of the "patch" embeddings is *always* better!

        :param img: Processed batch of images :: [bsz, 3, 224, 224]
        :param mode: Type of representation to extract -- `patch` (sequence of patch embeddings) or `cls` (<CLS>)

        :return: Extracted representations given img input.
        """
        assert img.ndim == 4, "Invalid input to `get_representations()`"
        assert mode in {"patch", "cls"}, f"Extraction mode `{mode}` not supported!"

        # Extract desired representations
        representations = self.encode(img)
        return representations[:, 1:] if mode == "patch" else representations[:, :1]

    def encode(self, img: torch.Tensor) -> torch.Tensor:
        """Run a single image through the MAE and extract patch embeddings."""

        # Note: All of this code is taken near-verbatim from the MVP repository...
        #   > Ref: https://github.com/ir413/mvp/blob/master/mvp/backbones/vit.py#L30
        patches = self.patch2embed(img)
        patches_pe = patches + self.encoder_pe[:, 1:, :]

        # Add CLS Token
        cls_token = self.cls_token + self.encoder_pe[:, :1, :]
        cls_tokens = cls_token.expand(img.shape[0], -1, -1)
        cls_patches = torch.cat([cls_tokens, patches_pe], dim=1)

        # Apply Transformer Blocks...
        for block in self.encoder_blocks:
            cls_patches = block(cls_patches)
        cls_patches = self.encoder_norm(cls_patches)
        return cls_patches

    def forward_encoder(
        self, imgs: torch.Tensor, mask_ratio: Optional[float] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        # Patchify + Position Embedding (without the CLS Token)
        patches = self.patch2embed(imgs)
        patches_pe = patches + self.encoder_pe[:, 1:, :]

        # Create mask (and go ahead and mask out patches at the same time)
        visible_patches, mask, restore_idxs = self.mask(patches_pe, mask_ratio)

        # Add the CLS Token
        cls_token = self.cls_token + self.encoder_pe[:, :1, :]
        cls_tokens = cls_token.expand(imgs.shape[0], -1, -1)
        cls_visible_patches = torch.cat([cls_tokens, visible_patches], dim=1)

        # Apply Transformer Blocks...
        for block in self.encoder_blocks:
            cls_visible_patches = block(cls_visible_patches)
        cls_visible_patches = self.encoder_norm(cls_visible_patches)

        return cls_visible_patches, mask, restore_idxs

    def forward_decoder(self, visible_patches: torch.Tensor, restore_idxs: torch.Tensor) -> torch.Tensor:
        # Project patches into decoder embedding dimension
        projected_patches = self.encoder2decoder(visible_patches)

        # Add Mask Tokens to Sequence
        mask_tokens = self.mask_token.repeat(
            projected_patches.shape[0], restore_idxs.shape[1] - visible_patches.shape[1], 1
        )

        # Remove & add back CLS Token as part of the "unshuffling"
        concatenated_patches = torch.cat([projected_patches[:, 1:, :], mask_tokens], dim=1)  # Skip CLS Token
        unshuffled_patches = torch.gather(
            concatenated_patches, dim=1, index=restore_idxs[..., None].repeat(1, 1, projected_patches.shape[2])
        )
        cls_unshuffled_patches = torch.cat([projected_patches[:, :1, :], unshuffled_patches], dim=1)  # Add CLS Token

        # Add Position Embeddings
        cls_decoder_patches = cls_unshuffled_patches + self.decoder_pe

        # Apply Transformer Blocks...
        for block in self.decoder_blocks:
            cls_decoder_patches = block(cls_decoder_patches)
        cls_decoder_patches = self.decoder_norm(cls_decoder_patches)

        # Run final projection, remove the CLS token, and return
        cls_decoder_prediction = self.decoder_prediction(cls_decoder_patches)
        decoder_prediction = cls_decoder_prediction[:, 1:, :]
        return decoder_prediction

    def patchify(self, imgs: torch.Tensor) -> torch.Tensor:
        """Convert a batch of images to their patched equivalents, by naive reshaping"""
        return rearrange(
            imgs,
            "bsz c (height patch_h) (width patch_w) -> bsz (height width) (patch_h patch_w c)",
            patch_h=self.patch_size,
            patch_w=self.patch_size,
        )

    def compute_loss(self, imgs: torch.Tensor, reconstructions: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        assert self.norm_pixel_loss, "`norm_pixel_loss` should always be true... false only for visualizations!"
        targets = self.patchify(imgs)

        # Normalize targets...
        mu, var = targets.mean(dim=-1, keepdim=True), targets.var(dim=-1, unbiased=True, keepdim=True)
        targets = (targets - mu) / ((var + 1e-6) ** 0.5)

        # Compute mean loss per patch first...
        mse = (reconstructions - targets) ** 2
        avg_loss_per_patch = mse.mean(dim=-1)

        # Compute mean loss only on *removed* patches and return
        return (avg_loss_per_patch * mask).sum() / mask.sum()

    def forward(
        self, imgs: torch.Tensor, mask_ratio: Optional[float] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        visible_patches, mask, restore_idxs = self.forward_encoder(imgs, mask_ratio)
        reconstructions = self.forward_decoder(visible_patches, restore_idxs)
        loss = self.compute_loss(imgs, reconstructions, mask)

        return loss, reconstructions, mask

    def configure_optimizer(self) -> Tuple[torch.optim.Optimizer, Callable[[int, float], float]]:
        # Short-Circuit on Valid Optimizers
        if self.optimizer not in ["adamw"]:
            raise NotImplementedError(f"Optimizer `{self.optimizer}` not supported - try [`adamw`] instead!")

        # Create Parameter Groups --> Bias terms, Normalization layer parameters shouldn't be decayed...
        #   > This is a compact rewrite of `param_groups_weight_decay()` from TIMM because I don't want the dependency
        decay, no_decay = [], []
        for name, param in self.named_parameters():
            if not param.requires_grad:
                continue

            # Check on any parameters with fewer than 2 dimensions or with "bias" in the name...
            if param.ndim <= 1 or name.endswith(".bias"):
                no_decay.append(param)
            else:
                decay.append(param)

        # Build Parameter Groups
        groups = [{"params": decay, "weight_decay": self.weight_decay}, {"params": no_decay, "weight_decay": 0.0}]

        # Compute LR -- MAE uses the `linear scaling rule` :: lr = base_lr * (effective_bsz / 256)
        #   > https://github.com/facebookresearch/mae/blob/main/PRETRAIN.md
        self.lr = self.base_lr * (self.effective_bsz / 256)

        # Create Optimizer & LR Scheduler
        optimizer = torch.optim.AdamW(groups, lr=self.lr, betas=self.betas)
        update_lr = get_lr_update(optimizer, self.schedule, self.lr, self.min_lr, self.warmup_epochs, self.max_epochs)
        return optimizer, update_lr



"""
extraction.py

General Extraction module definitions & associated utilities.

References:
    - Set Transformers (MAP): https://arxiv.org/abs/1810.00825.pdf
"""
from typing import Callable

import torch
import torch.nn as nn
from einops import repeat

from voltron.models.util.transformer import RMSNorm, SwishGLU

# === Multiheaded Attention Pooling ===


# As defined in Set Transformers () -- basically the above, additionally taking in
# a set of $k$ learned "seed vectors" that are used to "pool" information.
class MAPAttention(nn.Module):
    def __init__(self, embed_dim: int, n_heads: int) -> None:
        """Multi-Input Multi-Headed Attention Operation"""
        super().__init__()
        assert embed_dim % n_heads == 0, "`embed_dim` must be divisible by `n_heads`!"
        self.n_heads, self.scale = n_heads, (embed_dim // n_heads) ** -0.5

        # Projections (no bias) --> separate for Q (seed vector), and KV ("pool" inputs)
        self.q, self.kv = nn.Linear(embed_dim, embed_dim, bias=False), nn.Linear(embed_dim, 2 * embed_dim, bias=False)
        self.proj = nn.Linear(embed_dim, embed_dim)

    def forward(self, seed: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        (B_s, K, C_s), (B_x, N, C_x) = seed.shape, x.shape
        assert C_s == C_x, "Seed vectors and pool inputs must have the same embedding dimensionality!"

        # Project Seed Vectors to `queries`
        q = self.q(seed).reshape(B_s, K, self.n_heads, C_s // self.n_heads).permute(0, 2, 1, 3)
        kv = self.kv(x).reshape(B_x, N, 2, self.n_heads, C_x // self.n_heads).permute(2, 0, 3, 1, 4)
        k, v = kv.unbind(0)

        # Attention --> compute weighted sum over values!
        scores = q @ (k.transpose(-2, -1) * self.scale)
        attn = scores.softmax(dim=-1)
        vals = (attn @ v).transpose(1, 2).reshape(B_s, K, C_s)

        # Project back to `embed_dim`
        return self.proj(vals)


class MAPBlock(nn.Module):
    def __init__(
        self,
        n_latents: int,
        embed_dim: int,
        n_heads: int,
        mlp_ratio: float = 4.0,
        do_rms_norm: bool = True,
        do_swish_glu: bool = True,
    ) -> None:
        """Multiheaded Attention Pooling Block -- note that for MAP, we adopt earlier post-norm conventions."""
        super().__init__()
        self.n_latents, self.embed_dim, self.n_heads = n_latents, embed_dim, 2 * n_heads

        # Projection Operator
        self.projection = nn.Linear(embed_dim, self.embed_dim)

        # Initialize Latents
        self.latents = nn.Parameter(torch.zeros(self.n_latents, self.embed_dim))
        nn.init.normal_(self.latents, std=0.02)

        # Custom MAP Attention (seed, encoder outputs) -> seed
        self.attn_norm = RMSNorm(self.embed_dim) if do_rms_norm else nn.LayerNorm(self.embed_dim, eps=1e-6)
        self.attn = MAPAttention(self.embed_dim, n_heads=self.n_heads)

        # Position-wise Feed-Forward Components
        self.mlp_norm = RMSNorm(self.embed_dim) if do_rms_norm else nn.LayerNorm(self.embed_dim, eps=1e-6)
        self.mlp = nn.Sequential(
            # Handle SwishGLU vs. GELU MLP...
            (
                SwishGLU(self.embed_dim, int(mlp_ratio * self.embed_dim))
                if do_swish_glu
                else nn.Sequential(nn.Linear(self.embed_dim, int(mlp_ratio * self.embed_dim)), nn.GELU())
            ),
            nn.Linear(int(mlp_ratio * self.embed_dim), self.embed_dim),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        latents = repeat(self.latents, "n_latents d -> bsz n_latents d", bsz=x.shape[0])
        latents = self.attn_norm(latents + self.attn(latents, self.projection(x)))
        latents = self.mlp_norm(latents + self.mlp(latents))
        return latents.squeeze(dim=1)


# MAP Extractor Instantiation --> factory for creating extractors with the given parameters.
def instantiate_extractor(backbone: nn.Module, n_latents: int = 1) -> Callable[[], nn.Module]:
    def initialize() -> nn.Module:
        return MAPBlock(n_latents, backbone.embed_dim, backbone.n_heads)

    return initialize



"""
vcond.py

PyTorch Module defining the Voltron `V-Cond` variant (single-frame with language-conditioning). In general, follows the
MAE recipe, with the architectural modifications described in the paper:
    - RMSNorm, for stability/performance ("Do Transformer Modifications Transfer...")
    - SwishGLU activations in the Attention Block Feed-Forward MLP (gated linear units) as used in PaLM
    - LayerScale with a default value of 0.1 (from Mistral/CaIT)

References:
    - https://github.com/facebookresearch/mae
    - https://github.com/lucidrains/x-transformers
"""
from typing import Callable, List, Optional, Tuple, Union

import torch
import torch.nn as nn
import transformers
from einops import rearrange, repeat

from voltron.models.util.optimization import get_lr_update
from voltron.models.util.transformer import Block, PatchEmbed, RMSNorm, get_2D_position_embeddings

# Suppress Transformers Logging
transformers.logging.set_verbosity_error()


class VCond(nn.Module):
    def __init__(
        self,
        resolution: int,
        patch_size: int,
        encoder_depth: int,
        encoder_embed_dim: int,
        encoder_n_heads: int,
        decoder_depth: int,
        decoder_embed_dim: int,
        decoder_n_heads: int,
        language_model: str,
        hf_cache: str,
        language_dim: int,
        optimizer: str,
        schedule: str,
        base_lr: float,
        min_lr: float,
        effective_bsz: float,
        betas: Tuple[float, float],
        weight_decay: float,
        warmup_epochs: int,
        max_epochs: int,
        mask_ratio: float = 0.75,
        mlp_ratio: float = 4.0,
        in_channels: int = 3,
        norm_pixel_loss: bool = True,
    ) -> None:
        """
        Initialize a VCond model with the requisite architecture parameters.

        :param resolution: Base image resolution -- usually 224 (ImageNet size).
        :param patch_size: Height/Width of each patch in pixels -- usually 16.
        :param encoder_depth: Number of Transformer blocks in the encoder -- should be greater than decoder.
        :param encoder_embed_dim: Core embedding/hidden dimension for encoder vision transformer backbone.
        :param encoder_n_heads: Number of heads for encoder multi-headed self-attention.
        :param decoder_depth: Number of Transformer blocks in the decoder -- should be relatively shallow.
        :param decoder_embed_dim: Core embedding/hidden dimension for encoder vision transformer backbone.
        :param decoder_n_heads: Number of heads for encoder multi-headed self-attention.
        :param language_model: Language model to freeze for encoding narrations/utterances.
        :param hf_cache: Cache directory to store pretrained models, for safe distributed training.
        :param language_dim: Dimensionality of the language embedding coming out of the pretrained LM.
        :param optimizer: String denoting which optimizer to use (for MAEs, usually `adamw`)
        :param schedule: Learning rate schedule to use; for Transformers a linear warmup + decay is recommended!
        :param base_lr: Base learning rate, to be scaled via a linear scaling rule (from scaling laws).
        :param min_lr: Minimum learning rate to decay to over the course of learning (usually 0.0)
        :param effective_bsz: Global batch size for update, dictates the scaling of the base_lr.
        :param betas: Adam optimizer betas (only applicable for `adam` and `adamw`. Prevents early loss spiking.
        :param weight_decay: Weight decay for global weight regularization (only applied to non-bias, non-LN layers).
        :param warmup_epochs: Number of epochs to warmup learning rate for linear warmup schedule.
        :param max_epochs: Total number of training epochs to be run.
        :param mask_ratio: Ratio for number of patches to mask out for MAE -- should be fairly high!
        :param mlp_ratio: Ratio for embedding size to Position-wise Feed-Forward MLP (gets shrunk back down).
        :param in_channels: Default number of channels in the base image -- almost always 3.
        :param norm_pixel_loss: Normalize decoder pixel targets for reconstruction (better perf, not interpretable).
        """
        super().__init__()
        self.resolution, self.patch_size, self.mask_ratio = resolution, patch_size, mask_ratio
        self.in_channels, self.norm_pixel_loss, self.mlp_ratio = in_channels, norm_pixel_loss, mlp_ratio
        self.optimizer, self.schedule, self.betas, self.weight_decay = optimizer, schedule, betas, weight_decay
        self.lr, self.base_lr, self.min_lr, self.effective_bsz = None, base_lr, min_lr, effective_bsz
        self.warmup_epochs, self.max_epochs = warmup_epochs, max_epochs
        self.language_dim = language_dim

        # Encoder/Decoder Parameters
        self.encoder_depth, self.decoder_depth = encoder_depth, decoder_depth
        self.encoder_embed_dim, self.encoder_n_heads = encoder_embed_dim, encoder_n_heads
        self.decoder_embed_dim, self.decoder_n_heads = decoder_embed_dim, decoder_n_heads

        # General Parameters (for downstream adaptation)
        self.embed_dim, self.n_heads = self.encoder_embed_dim, self.encoder_n_heads

        # MAE Encoder Parameters --> No CLS Token!
        self.patch2embed = PatchEmbed(
            self.resolution, self.patch_size, self.encoder_embed_dim, in_channels=self.in_channels
        )
        self.encoder_pe = nn.Parameter(
            torch.zeros(1, self.patch2embed.num_patches, self.encoder_embed_dim), requires_grad=False
        )
        self.encoder_blocks = nn.ModuleList(
            [
                Block(
                    self.encoder_embed_dim,
                    self.encoder_n_heads,
                    self.mlp_ratio,
                    do_rms_norm=True,
                    do_swish_glu=True,
                    do_layer_scale=True,
                )
                for _ in range(self.encoder_depth)
            ]
        )
        self.encoder_norm = RMSNorm(self.encoder_embed_dim)

        # Projection from Language Embedding to Decoder
        self.lang2encoder = nn.Linear(self.language_dim, self.encoder_embed_dim)

        # Projection from Encoder to Decoder
        self.encoder2decoder = nn.Linear(self.encoder_embed_dim, self.decoder_embed_dim)

        # MAE Decoder Parameters -- Remember the MASK Token!
        self.mask_token = nn.Parameter(torch.zeros(1, 1, self.decoder_embed_dim))
        self.decoder_pe = nn.Parameter(
            torch.zeros(1, self.patch2embed.num_patches, self.decoder_embed_dim), requires_grad=False
        )
        self.decoder_blocks = nn.ModuleList(
            [
                Block(
                    self.decoder_embed_dim,
                    self.decoder_n_heads,
                    self.mlp_ratio,
                    do_rms_norm=True,
                    do_swish_glu=True,
                    do_layer_scale=True,
                )
                for _ in range(self.decoder_depth)
            ]
        )
        self.decoder_norm = RMSNorm(self.decoder_embed_dim)
        self.decoder_prediction = nn.Linear(self.decoder_embed_dim, (patch_size**2) * in_channels, bias=True)

        # VCond -- Add "Image" and "Language" Modifier Tokens...
        self.img_token = nn.Parameter(torch.zeros(1, 1, self.encoder_embed_dim))
        self.lang_token = nn.Parameter(torch.zeros(1, 1, self.encoder_embed_dim))

        # Initialize all Weights
        self.initialize_weights()

        # @AFTER INITIALIZATION -- Create Language Model & Language Reward MLP --> LM has requires_grad = False
        #   > For BERT models, our "embedding" is just going to be the last hidden state
        #   > Assumes inputs to forward pass are pre-tokenized!
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(language_model, cache_dir=hf_cache)
        self.lm = transformers.AutoModel.from_pretrained(language_model, cache_dir=hf_cache)
        self.lm.eval()

        # Shape Assertion -- make sure self.language_dim actually is the same as the LM dimension!
        assert self.lm.config.dim == self.language_dim, "Language model embedding dimension != self.language_dim!"

        # Freeze the LM
        for _, param in self.lm.named_parameters():
            param.requires_grad = False

    def initialize_weights(self) -> None:
        # Position Encoding -- Fixed 2D Sine-Cosine Embeddings
        enc_pe = get_2D_position_embeddings(self.encoder_embed_dim, int(self.patch2embed.num_patches**0.5))
        self.encoder_pe.data.copy_(torch.from_numpy(enc_pe).float().unsqueeze(0))
        dec_pe = get_2D_position_embeddings(self.decoder_embed_dim, int(self.patch2embed.num_patches**0.5))
        self.decoder_pe.data.copy_(torch.from_numpy(dec_pe).float().unsqueeze(0))

        # Initialize PatchEmbedding as a Linear...
        nn.init.xavier_uniform_(self.patch2embed.proj.weight.data.view([self.patch2embed.proj.weight.data.shape[0], -1]))

        # Initialize Mask Token, Img Token, Lang Token w/ Truncated Normal
        nn.init.normal_(self.mask_token, std=0.02)
        nn.init.normal_(self.img_token, std=0.02)
        nn.init.normal_(self.lang_token, std=0.02)

        # Default Transformer initializer on everything else...
        self.apply(self.transformer_initializer)

    @staticmethod
    def transformer_initializer(m: nn.Module) -> None:
        # Use `xavier_uniform` following Jax ViT
        if isinstance(m, nn.Linear):
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0.0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.weight, 1.0)
            nn.init.constant_(m.bias, 0.0)

    def encode_language(self, lang: torch.Tensor, lang_mask: torch.Tensor) -> torch.Tensor:
        """Encode language by feeding the *pre-tokenized text* through the frozen language model."""
        self.lm.eval()
        with torch.no_grad():
            transformer_embeddings = self.lm(lang, attention_mask=lang_mask).last_hidden_state
        return transformer_embeddings

    def mask(
        self, patches: torch.Tensor, mask_ratio: Optional[float] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Perform per-sample random masking by shuffling :: uses argsort random noise to identify masked patches."""
        bsz, n_patches, embed_dim = patches.shape
        if mask_ratio is not None:
            n_keep = int(n_patches * (1 - mask_ratio))
        else:
            n_keep = int(n_patches * (1 - self.mask_ratio))

        # Sample noise of n_patches size, argsort to get shuffled IDs, argsort again to get "unshuffle"
        #   > For clarity -- argsort is an invertible transformation (if argsort `restore`, recovers `shuffle`)
        shuffle_idxs = torch.argsort(torch.rand(bsz, n_patches, device=patches.device), dim=1)
        restore_idxs = torch.argsort(shuffle_idxs, dim=1)

        # Get "keep" (visible) patches
        visible_patches = torch.gather(patches, dim=1, index=shuffle_idxs[:, :n_keep, None].repeat(1, 1, embed_dim))

        # Generate the binary mask --> IMPORTANT :: `0` is keep, `1` is remove (following MAE convention)
        mask = torch.ones(bsz, n_patches, device=patches.device)
        mask[:, :n_keep] = 0
        mask = torch.gather(mask, dim=1, index=restore_idxs)

        return visible_patches, mask, restore_idxs

    def get_representations(
        self, img: torch.Tensor, language: Optional[Union[List[str], Tuple[str]]] = None, mode: str = "multimodal"
    ) -> torch.Tensor:
        """
        Given either a singleton (img, language) pair or a batch of images and language, extract representations
        subject to the specified mode in < multimodal | visual >.

        :param img: Processed batch of images :: [bsz, 3, 224, 224]
        :param language: Input language as `List[str] | Tuple[str] | None`
        :param mode: Type of representations to extract -- `multimodal` (both vision+text), `visual` (visual only)

        :return: Extracted representations given (img, language) input as sequence.
        """
        assert img.ndim == 4 and (
            language is None or isinstance(language, list) or isinstance(language, tuple)
        ), "Invalid input to `get_representations()`"
        assert mode in {"multimodal", "visual"}, f"Extraction mode `{mode}` not supported!"

        # Tokenize Language --> note max length is 20!
        if language is None:
            lang, lang_mask = [torch.zeros(img.shape[0], 20, dtype=int, device=self.lm.device) for _ in range(2)]
            lang[:, 0], lang_mask[:, 0] = self.tokenizer.cls_token_id, 1
        else:
            tokens = self.tokenizer(language, return_tensors="pt", max_length=20, padding="max_length", truncation=True)
            lang, lang_mask = tokens["input_ids"].to(self.lm.device), tokens["attention_mask"].to(self.lm.device)

            # Tile Language & Language Mask if mismatch with # images!
            if not len(lang) == len(img):
                lang = repeat(lang, "b seq -> (bsz b) seq", bsz=img.size(0))
                lang_mask = repeat(lang_mask, "b seq -> (bsz b) seq", bsz=img.size(0))

        # Extract desired representations...
        representations = self.encode(img, lang, lang_mask)
        return representations if mode == "multimodal" else representations[:, : -lang_mask.shape[-1]]

    def encode(self, img: torch.Tensor, lang: torch.Tensor, lang_mask: torch.Tensor) -> torch.Tensor:
        """Default representation extraction function, given a batch of images and tokenized language."""
        lang_embeddings = self.encode_language(lang, lang_mask)
        projected_language = self.lang2encoder(lang_embeddings)

        # Patchify + Position Embedding
        patches = self.patch2embed(img)
        patches_pe = patches + self.encoder_pe

        # Add "modality" embeddings to patches & language
        img_embeddings, lang_embeddings = patches_pe + self.img_token, projected_language + self.lang_token

        # Create "dummy" visible mask, concatenate image patches & language, feed to Transformer
        patches_mask = torch.ones_like(img_embeddings[..., -1], dtype=lang_mask.dtype)
        multimodal_embeddings = torch.cat([img_embeddings, lang_embeddings], dim=1)  # Merge on sequence length...
        multimodal_mask = torch.cat([patches_mask, lang_mask], dim=1)  # Merge on sequence length...

        # Apply Transformer Blocks...
        for block in self.encoder_blocks:
            multimodal_embeddings = block(multimodal_embeddings, multimodal_mask)
        multimodal_embeddings = self.encoder_norm(multimodal_embeddings)

        # Return the full sequence of multimodal embeddings...
        return multimodal_embeddings

    def forward_encoder(
        self, img: torch.Tensor, lang: torch.Tensor, lang_mask: torch.Tensor, mask_ratio: Optional[float] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        # Patchify + Position Embedding
        patches = self.patch2embed(img)
        patches_pe = patches + self.encoder_pe

        # Create mask (and go ahead and mask out patches at the same time)
        visible_patches, mask, restore_idxs = self.mask(patches_pe, mask_ratio)

        # Add "modality" embeddings to patches & language
        visible_patches, lang = visible_patches + self.img_token, lang + self.lang_token

        # Create "dummy" visible mask, concatenate image patches & language, feed to Transformer
        visible_mask = torch.ones_like(visible_patches[..., -1], dtype=lang_mask.dtype)
        multimodal_embedding = torch.cat([visible_patches, lang], dim=1)  # Merge on sequence length...
        multimodal_mask = torch.cat([visible_mask, lang_mask], dim=1)  # Merge on sequence length...

        # Apply Transformer Blocks...
        for block in self.encoder_blocks:
            multimodal_embedding = block(multimodal_embedding, multimodal_mask)
        multimodal_embedding = self.encoder_norm(multimodal_embedding)

        # Split multimodal embedding, remove language and return only the visible patches!
        visible_patches = multimodal_embedding[:, : -lang_mask.shape[-1], ...]
        return visible_patches, mask, restore_idxs

    def forward_decoder(self, visible_patches: torch.Tensor, restore_idxs: torch.Tensor) -> torch.Tensor:
        # Project patches into decoder embedding dimension
        projected_patches = self.encoder2decoder(visible_patches)

        # Add Mask Tokens to Sequence & Unshuffle
        mask_tokens = self.mask_token.repeat(
            projected_patches.shape[0], restore_idxs.shape[1] - visible_patches.shape[1], 1
        )
        concatenated_patches = torch.cat([projected_patches, mask_tokens], dim=1)
        unshuffled_patches = torch.gather(
            concatenated_patches, dim=1, index=restore_idxs[..., None].repeat(1, 1, self.decoder_embed_dim)
        )

        # Add Position Embeddings
        decoder_patches = unshuffled_patches + self.decoder_pe

        # Apply Transformer Blocks...
        for block in self.decoder_blocks:
            decoder_patches = block(decoder_patches)
        decoder_patches = self.decoder_norm(decoder_patches)

        # Run final projection & return
        return self.decoder_prediction(decoder_patches)

    def patchify(self, imgs: torch.Tensor) -> torch.Tensor:
        """Convert a batch of images to their patched equivalents, by naive reshaping"""
        return rearrange(
            imgs,
            "bsz c (height patch_h) (width patch_w) -> bsz (height width) (patch_h patch_w c)",
            patch_h=self.patch_size,
            patch_w=self.patch_size,
        )

    def compute_loss(self, imgs: torch.Tensor, reconstructions: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        assert self.norm_pixel_loss, "`norm_pixel_loss` should always be true... false only for visualizations!"
        targets = self.patchify(imgs)

        # Normalize targets...
        mu, var = targets.mean(dim=-1, keepdim=True), targets.var(dim=-1, unbiased=True, keepdim=True)
        targets = (targets - mu) / ((var + 1e-6) ** 0.5)

        # Compute mean loss per patch first...
        mse = (reconstructions - targets) ** 2
        avg_loss_per_patch = mse.mean(dim=-1)

        # Compute mean loss only on *removed* patches and return
        return (avg_loss_per_patch * mask).sum() / mask.sum()

    def forward(
        self, img: torch.Tensor, lang: torch.Tensor, lang_mask: torch.Tensor, mask_ratio: Optional[float] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        # First, get precomputed language embeddings & project to encoder_embed_dim; language only used in encoder!
        lang_embeddings = self.encode_language(lang, lang_mask)
        projected_language = self.lang2encoder(lang_embeddings)

        # MAE --> does image masking in the body of the model...
        visible_patches, mask, restore_idxs = self.forward_encoder(img, projected_language, lang_mask, mask_ratio)
        reconstructions = self.forward_decoder(visible_patches, restore_idxs)
        loss = self.compute_loss(img, reconstructions, mask)
        return loss, reconstructions, mask

    def configure_optimizer(self) -> Tuple[torch.optim.Optimizer, Callable[[int, float], float]]:
        # Short-circuit on valid optimizers
        if self.optimizer not in ["adamw"]:
            raise NotImplementedError(f"Optimizer `{self.optimizer}` not supported - try [`adamw`] instead!")

        # Create Parameter Groups --> bias terms, normalization layer parameters shouldn't be decayed...
        #   > This is a compact rewrite of `param_groups_weight_decay()` from TIMM because I don't want the dependency
        decay, no_decay = [], []
        for name, param in self.named_parameters():
            if not param.requires_grad:
                continue

            # Check on any parameters with fewer than 2 dimensions or with "bias" in the name...
            if param.ndim <= 1 or name.endswith(".bias"):
                no_decay.append(param)
            else:
                decay.append(param)

        # Build Parameter Groups
        groups = [{"params": decay, "weight_decay": self.weight_decay}, {"params": no_decay, "weight_decay": 0.0}]

        # Compute LR -- MAE uses the `linear scaling rule` :: lr = base_lr * (effective_bsz / 256)
        #   > https://github.com/facebookresearch/mae/blob/main/PRETRAIN.md
        self.lr = self.base_lr * (self.effective_bsz / 256)

        # Create Optimizer & LR Scheduler
        optimizer = torch.optim.AdamW(groups, lr=self.lr, betas=self.betas)
        update_lr = get_lr_update(optimizer, self.schedule, self.lr, self.min_lr, self.warmup_epochs, self.max_epochs)
        return optimizer, update_lr



"""
transformer.py

General Transformer modules & utilities.

References:
    - https://github.com/facebookresearch/mae
    - https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py
"""
from typing import Optional

import numpy as np
import torch
import torch.nn as nn
from einops import rearrange

# === Position Encoding Utilities ===


# Helper/Utility Function -- computes simple 1D sinusoidal position embeddings for both 1D/2D use cases.
#   > We'll be combining two 1D sin-cos (traditional) position encodings for height/width of an image (grid features).
def get_1D_sine_cosine(dim: int, pos: np.ndarray) -> np.ndarray:
    omega = np.arange(dim // 2, dtype=np.float32) / (dim / 2.0)
    omega = 1.0 / (10000**omega)
    out = np.einsum("m,d->md", pos.reshape(-1), omega)  # [flatten(pos) x omega] -- outer product!
    emb_sin, emb_cos = np.sin(out), np.cos(out)
    return np.concatenate([emb_sin, emb_cos], axis=1)  # [flatten(pos) x D]


# 1D Sine-Cosine Position Embedding -- standard from "Attention is all you need!"
def get_1D_position_embeddings(embed_dim: int, length: int) -> np.ndarray:
    return get_1D_sine_cosine(embed_dim, np.arange(length))


# 2D Sine-Cosine Position Embedding (from MAE repository)
#   > https://github.com/facebookresearch/mae/blob/efb2a8062c206524e35e47d04501ed4f544c0ae8/util/pos_embed.py#L20
def get_2D_position_embeddings(embed_dim: int, grid_size: int, cls_token: bool = False) -> np.ndarray:
    # Create 2D Position embeddings by taking cross product of height and width and splicing 1D embeddings...
    grid_h, grid_w = np.arange(grid_size, dtype=np.float32), np.arange(grid_size, dtype=np.float32)
    grid = np.stack(np.meshgrid(grid_w, grid_h), axis=0).reshape(2, 1, grid_size, grid_size)  # w goes first?

    # Use half of dimensions to encode grid_h, other half to encode grid_w
    emb_h, emb_w = get_1D_sine_cosine(embed_dim // 2, grid[0]), get_1D_sine_cosine(embed_dim // 2, grid[1])
    pos_embed = np.concatenate([emb_h, emb_w], axis=1)

    # CLS token handling (only for R-MVP)
    if cls_token:
        pos_embed = np.concatenate([np.zeros([1, embed_dim]), pos_embed], axis=0)

    return pos_embed


# === Vision Transformer Building Blocks ===


# Patch Embedding Module
class PatchEmbed(nn.Module):
    def __init__(
        self,
        resolution: int,
        patch_size: int,
        embed_dim: int,
        in_channels: int = 3,
        flatten: bool = True,
    ):
        super().__init__()
        self.resolution, self.patch_size = (resolution, resolution), (patch_size, patch_size)
        self.grid_size = (self.resolution[0] // self.patch_size[0], self.resolution[1] // self.patch_size[1])
        self.num_patches = self.grid_size[0] * self.grid_size[1]
        self.flatten = flatten
        self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=self.patch_size, stride=self.patch_size)

    def forward(self, patches: torch.Tensor) -> torch.Tensor:
        patch_embeddings = self.proj(patches)
        if self.flatten:
            return rearrange(patch_embeddings, "bsz embed patch_h patch_w -> bsz (patch_h patch_w) embed")
        return patch_embeddings


# === Stability Utilities ===


# LayerScale -- Trainable scaling for residual blocks -- Mistral/CaIT
class LayerScale(nn.Module):
    def __init__(self, dim: int, init_values: float = 0.1) -> None:  # CaIT :: 0.1 -> lay 12, 1e-5 -> lay 24, 1e-6...
        super().__init__()
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.gamma


# RMSNorm -- Better, simpler alternative to LayerNorm
class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-8) -> None:
        super().__init__()
        self.scale, self.eps = dim**-0.5, eps
        self.g = nn.Parameter(torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        norm = torch.norm(x, dim=-1, keepdim=True) * self.scale
        return x / norm.clamp(min=self.eps) * self.g


# SwishGLU -- A Gated Linear Unit (GLU) with the Swish activation; always better than GELU MLP!
class SwishGLU(nn.Module):
    def __init__(self, in_dim: int, out_dim: int) -> None:
        super().__init__()
        self.act, self.project = nn.SiLU(), nn.Linear(in_dim, 2 * out_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        projected, gate = self.project(x).tensor_split(2, dim=-1)
        return projected * self.act(gate)


# === Fundamental Transformer Building Blocks ===


class Attention(nn.Module):
    def __init__(self, embed_dim: int, n_heads: int, dropout: float = 0.0) -> None:
        """Multi-Headed Self-Attention Operation"""
        super().__init__()
        assert embed_dim % n_heads == 0, "`embed_dim` must be divisible by `n_heads`!"
        self.n_heads, self.scale = n_heads, (embed_dim // n_heads) ** -0.5
        self.attn_softmax = None

        # Projections
        self.qkv, self.proj = nn.Linear(embed_dim, 3 * embed_dim, bias=True), nn.Linear(embed_dim, embed_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        B, N, C = x.shape

        # Project to Q-K-V
        qkv = self.qkv(x).reshape(B, N, 3, self.n_heads, C // self.n_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)

        # Self-attention -- with masking!
        scores = q @ (k.transpose(-2, -1) * self.scale)
        if mask is not None:
            if mask.ndim == 2:
                mask = rearrange(mask, "bsz seq -> bsz 1 seq 1")
            elif mask.ndim != 4:
                raise NotImplementedError("Attention got `mask` of shape not in {2, 4}!")

            # Mask out by filling indices with negative infinity...
            scores = scores.masked_fill(mask == 0, torch.finfo(scores.dtype).min)

        # Compute weighted sum over values
        self.attn_softmax = scores.softmax(dim=-1)
        vals = (self.attn_softmax @ v).transpose(1, 2).reshape(B, N, C)

        # Project back to `embed_dim` -- with optional dropout
        vals = self.dropout(self.proj(vals))
        return vals


class Block(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        n_heads: int,
        mlp_ratio: float = 4.0,
        dropout: float = 0.0,
        do_rms_norm: bool = False,
        do_swish_glu: bool = False,
        do_layer_scale: bool = False,
    ) -> None:
        """
        Transformer Block Implementation (modality-agnostic).

        :param embed_dim: Core embedding/hidden dimension for vision transformer backbone.
        :param n_heads: Number of heads for multi-headed self-attention.
        :param mlp_ratio: Ratio for embedding size to position-wise feed-forward MLP (gets shrunk back down).
        :param dropout: [Optional] dropout for projection layer and MLPs -- for MAEs, always 0.0!
        :param do_rms_norm: Boolean whether or not to use RMSNorm in lieu of LayerNorm within block.
        :param do_swish_glu: Use the Swish-variant of the Gated Linear Unit for the feed-forward layers.
        :param do_layer_scale: Boolean whether or not to use LayerScale from Mistral/CaIT w/ initialization of 0.1.
        """
        super().__init__()
        self.embed_dim, self.n_heads, self.do_layer_scale = embed_dim, n_heads, do_layer_scale

        # Attention Components
        self.pre_norm_attn = RMSNorm(self.embed_dim) if do_rms_norm else nn.LayerNorm(self.embed_dim, eps=1e-6)
        self.attn = Attention(self.embed_dim, n_heads=n_heads, dropout=dropout)
        if do_layer_scale:
            self.layer_scale_attn = LayerScale(self.embed_dim)

        # Position-wise Feed-Forward Components
        self.pre_norm_mlp = RMSNorm(self.embed_dim) if do_rms_norm else nn.LayerNorm(self.embed_dim, eps=1e-6)
        self.mlp = nn.Sequential(
            # Handle SwishGLU vs. GELU MLP...
            (
                SwishGLU(embed_dim, int(mlp_ratio * embed_dim))
                if do_swish_glu
                else nn.Sequential(nn.Linear(embed_dim, int(mlp_ratio * embed_dim)), nn.GELU())
            ),
            nn.Dropout(dropout),
            nn.Linear(int(mlp_ratio * embed_dim), embed_dim),
        )
        if self.do_layer_scale:
            self.layer_scale_mlp = LayerScale(self.embed_dim)

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        if self.do_layer_scale:
            x = x + self.layer_scale_attn(self.attn(self.pre_norm_attn(x), mask))
            x = x + self.layer_scale_mlp(self.mlp(self.pre_norm_mlp(x)))
        else:
            x = x + self.attn(self.pre_norm_attn(x), mask)
            x = x + self.mlp(self.pre_norm_mlp(x))
        return x



"""
vrn3m.py

PyTorch Module defining an R3M model (with a ResNet 50 encoder), exactly as described in Nair et. al. 2021, with all the
requisite hyperparameters.

Reference:
    - https://github.com/facebookresearch/r3m
"""
from typing import Callable, Tuple

import torch
import torch.nn as nn
import transformers
from einops import rearrange
from torchvision.models import resnet50

from voltron.models.util.optimization import get_lr_update

# Suppress Transformers Logging
transformers.logging.set_verbosity_error()


class VRN3M(nn.Module):
    def __init__(
        self,
        resolution: int,
        fc_dim: int,
        language_model: str,
        hf_cache: str,
        language_dim: int,
        reward_dim: int,
        n_negatives: int,
        lang_reward_weight: float,
        tcn_weight: float,
        l1_weight: float,
        l2_weight: float,
        optimizer: str,
        lr: float,
        eps: float = 1e-8,
    ):
        """
        Initialize an ResNet-50 R3M model with the required architecture parameters.

        :param resolution: Base image resolution -- usually 224 (ImageNet size).
        :param fc_dim: Dimensionality of the pooled embedding coming out of the ResNet (for RN50, fc_dim = 2048)
        :param language_model: Language model to freeze for encoding narrations/utterances.
        :param hf_cache: Cache directory to store pretrained models, for safe distributed training.
        :param language_dim: Dimensionality of the language embedding coming out of the pretrained LM.
        :param reward_dim: Hidden layer dimensionality for the language-reward MLP.
        :param n_negatives: Number of cross-batch negatives to sample for contrastive learning.
        :param lang_reward_weight: Weight applied to the contrastive "language alignment" loss term.
        :param tcn_weight: Weight applied to the time contrastive loss term.
        :param l1_weight: Weight applied to the L1 regularization loss term.
        :param l2_weight: Weight applied to the L2 regularization loss term.
        :param optimizer: String denoting which optimizer to use (for R3M, usually `adam`).
        :param lr: Learning rate (fixed for ResNet R3M models) for training.
        :param eps: Epsilon for preventing divide by zero in the InfoNCE loss terms.
        """
        super().__init__()
        self.resolution, self.fc_dim, self.n_negatives, self.eps = resolution, fc_dim, n_negatives, eps
        self.language_dim, self.reward_dim, self.optimizer, self.lr = language_dim, reward_dim, optimizer, lr
        self.embed_dim = self.fc_dim

        # Weights for each loss term
        self.lang_reward_weight, self.tcn_weight = lang_reward_weight, tcn_weight
        self.l1_weight, self.l2_weight = l1_weight, l2_weight

        # Create ResNet50 --> set `rn.fc` to the Identity() to extract final features of dim = `fc_dim`
        self.resnet = resnet50(weights=None)
        self.resnet.fc = nn.Identity()
        self.resnet.train()

        # Create Language Reward Model
        self.language_reward = nn.Sequential(
            nn.Linear(self.fc_dim + self.fc_dim + self.language_dim, self.reward_dim),
            nn.ReLU(),
            nn.Linear(self.reward_dim, self.reward_dim),
            nn.ReLU(),
            nn.Linear(self.reward_dim, self.reward_dim),
            nn.ReLU(),
            nn.Linear(self.reward_dim, self.reward_dim),
            nn.ReLU(),
            nn.Linear(self.reward_dim, 1),
            nn.Sigmoid(),
        )

        # Create Language Model & Language Reward MLP --> LM has requires_grad = False
        #   > For BERT models, our "embedding" is just going to be the last hidden state
        #   > Assumes inputs to forward pass are pre-tokenized!
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(language_model, cache_dir=hf_cache)
        self.lm = transformers.AutoModel.from_pretrained(language_model, cache_dir=hf_cache)
        self.lm.eval()

        # Shape Assertion -- make sure self.language_dim actually is the same as the LM dimension!
        assert self.lm.config.dim == self.language_dim, "Language model embedding dimension != self.language_dim!"

        # Freeze the LM
        for _name, param in self.lm.named_parameters():
            param.requires_grad = False

    def get_representations(self, img: torch.Tensor) -> torch.Tensor:
        """
        Given a single image, extract R3M "default" (ResNet pooled) dense representation.

        :param img: Processed batch of images :: [bsz, 3, 224, 224]
        :return: Extracted R3M dense representation given img input.
        """
        assert img.ndim == 4, "Invalid input to `get_representations()`"
        representation = self.resnet(img)
        return representation.unsqueeze(1)

    def encode_images(self, imgs: torch.Tensor) -> torch.Tensor:
        """Feed images through ResNet-50 to get single embedding after global average pooling."""
        return self.resnet(imgs)

    def encode_language(self, lang: torch.Tensor, lang_mask: torch.Tensor) -> torch.Tensor:
        """Encode language by feeding the *pre-tokenized text* through the frozen language model."""
        self.lm.eval()
        with torch.no_grad():
            transformer_embeddings = self.lm(lang, attention_mask=lang_mask).last_hidden_state
            return transformer_embeddings.mean(dim=1)

    def get_reward(self, initial: torch.Tensor, later: torch.Tensor, lang: torch.Tensor) -> torch.Tensor:
        return self.language_reward(torch.cat([initial, later, lang], dim=-1)).squeeze()

    def extract_features(self, img: torch.Tensor) -> torch.Tensor:
        """Run a single image of shape [1, 3, 224, 224] through the ResNet and extract the feature."""
        return self.encode_images(img).detach()

    def forward(self, imgs: torch.Tensor, lang: torch.Tensor, lang_mask: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """
        Run a forward pass through the model, computing the *full* R3M loss -- the TCN contrastive loss, the Language
        Alignment loss, and both sparsity losses, as well as the full loss (which will get optimized)!

        :param imgs: A [bsz, 5, in_channels, resolution, resolution] tensor of (start, i, j, k, end) sequences.
        :param lang: Tokenized language of dimensionality [bsz, seq_len] to be fed to the language model.
        :param lang_mask: Attention mask computed by the tokenizer, as a result of padding to the max_seq_len.

        :return: Tuple of losses, as follows:
            > (combined_loss, tcn_loss, reward_loss, l1_loss, l2_loss, tcn_acc, reward_acc)
        """
        # Encode each image separately... feed to transformer... then reshape
        all_images = rearrange(imgs, "bsz n_states c res1 res2 -> (bsz n_states) c res1 res2", n_states=5)
        all_embeddings = self.encode_images(all_images)
        initial, state_i, state_j, state_k, final = rearrange(
            all_embeddings, "(bsz n_states) embed -> n_states bsz embed", n_states=5
        )

        # Compute Regularization Losses
        l1_loss = torch.linalg.norm(all_embeddings, ord=1, dim=-1).mean()
        l2_loss = torch.linalg.norm(all_embeddings, ord=2, dim=-1).mean()

        # Compute TCN Loss
        tcn_loss, tcn_acc = self.get_time_contrastive_loss(state_i, state_j, state_k)

        # Compute Language Alignment/Predictive Loss
        lang_reward_loss, rew_acc = self.get_reward_loss(lang, lang_mask, initial, state_i, state_j, state_k, final)

        # Compute full weighted loss & return...
        loss = (
            (self.l1_weight * l1_loss)
            + (self.l2_weight * l2_loss)
            + (self.tcn_weight * tcn_loss)
            + (self.lang_reward_weight * lang_reward_loss)
        )
        return loss, tcn_loss, lang_reward_loss, l1_loss, l2_loss, tcn_acc, rew_acc

    @staticmethod
    def time_similarity(state_x: torch.Tensor, state_y: torch.Tensor, use_l2: bool = True) -> torch.Tensor:
        """Computes similarity between embeddings via -L2 distance."""
        assert use_l2, "Non-L2 time-similarity functions not yet implemented!"
        return -torch.linalg.norm(state_x - state_y, dim=-1)

    def get_time_contrastive_loss(
        self, state_i: torch.Tensor, state_j: torch.Tensor, state_k: torch.Tensor
    ) -> Tuple[torch.Tensor, ...]:
        """Evaluates the Time-Contrastive Loss, computed via InfoNCE."""

        # *Punchline* - we want `sim(i, j)` to be higher than `sim(i, k)` for some k > j (goes both ways)
        # `Reward(s*_0, s*_<t, l)` where * indicates a different video...
        #       > As our positive examples --> we sample (s_i, s_j) and (s_j, s_k).
        #       > Our negatives --> other pairs from the triplet, cross-batch negatives!
        sim_i_j_exp = torch.exp(self.time_similarity(state_i, state_j))
        sim_j_k_exp = torch.exp(self.time_similarity(state_j, state_k))

        # Add a "hard" negative!
        neg_i_k_exp = torch.exp(self.time_similarity(state_i, state_k))

        # Obtain *cross-batch* negatives
        bsz, neg_i, neg_j = state_i.shape[0], [], []
        for _ in range(self.n_negatives):
            neg_idx = torch.randperm(bsz)
            state_i_shuf = state_i[neg_idx]
            neg_idx = torch.randperm(bsz)
            state_j_shuf = state_j[neg_idx]
            neg_i.append(self.time_similarity(state_i, state_i_shuf))
            neg_j.append(self.time_similarity(state_j, state_j_shuf))
        neg_i_exp, neg_j_exp = torch.exp(torch.stack(neg_i, -1)), torch.exp(torch.stack(neg_j, -1))

        # Compute InfoNCE
        denominator_i = sim_i_j_exp + neg_i_k_exp + neg_i_exp.sum(-1)
        denominator_j = sim_j_k_exp + neg_i_k_exp + neg_j_exp.sum(-1)
        nce_i = -torch.log(self.eps + (sim_i_j_exp / (self.eps + denominator_i)))
        nce_j = -torch.log(self.eps + (sim_j_k_exp / (self.eps + denominator_j)))
        nce = (nce_i + nce_j) / 2

        # Compute "accuracy"
        i_j_acc = (1.0 * (sim_i_j_exp > neg_i_k_exp)).mean()
        j_k_acc = (1.0 * (sim_j_k_exp > neg_i_k_exp)).mean()
        acc = (i_j_acc + j_k_acc) / 2

        return nce.mean(), acc

    def get_reward_loss(
        self,
        lang: torch.Tensor,
        lang_mask: torch.Tensor,
        initial: torch.Tensor,
        state_i: torch.Tensor,
        state_j: torch.Tensor,
        state_k: torch.Tensor,
        final: torch.Tensor,
    ) -> Tuple[torch.Tensor, ...]:
        """Evaluates the Language-Alignment Reward Loss, computed via InfoNCE."""
        lang_embed = self.encode_language(lang, lang_mask)

        # *Punchline* - we want `Reward(s_0, s_t, l)` to be higher than `Reward(s_0, s_<t, l)` AND
        # `Reward(s*_0, s*_<t, l)` where * indicates a different video...
        #       > As our positive examples --> we sample s_j, s_k, and s_final (excluding s_i)
        pos_final_exp = torch.exp(self.get_reward(initial, final, lang_embed))
        pos_j_exp = torch.exp(self.get_reward(initial, state_j, lang_embed))
        pos_k_exp = torch.exp(self.get_reward(initial, state_k, lang_embed))

        # Add the within-context negatives <--> these are the most informative examples!
        #   > We use initial, initial as a negative for the first one, just to get reward model to "capture progress"
        negs_final = [self.get_reward(initial, initial, lang_embed)]
        negs_j = [self.get_reward(initial, state_i, lang_embed)]
        negs_k = [self.get_reward(initial, state_j, lang_embed)]

        # Cross Batch Negatives -- same as positives (indexing), but from a different batch!
        #   > @SK :: Unclear how well this will unroll on TPUs...
        bsz = initial.shape[0]
        for _ in range(self.n_negatives):
            # We get three random indices to further minimize correlation... from the R3M codebase!
            neg_idx = torch.randperm(bsz)
            negs_final.append(self.get_reward(initial[neg_idx], final[neg_idx], lang_embed))
            neg_idx = torch.randperm(bsz)
            negs_j.append(self.get_reward(initial[neg_idx], state_j[neg_idx], lang_embed))
            neg_idx = torch.randperm(bsz)
            negs_k.append(self.get_reward(initial[neg_idx], state_k[neg_idx], lang_embed))

        # Flatten & exponentiate; get ready for the InfoNCE
        negs_final, negs_j, negs_k = torch.stack(negs_final, -1), torch.stack(negs_j, -1), torch.stack(negs_k, -1)
        negs_final_exp, negs_j_exp, negs_k_exp = torch.exp(negs_final), torch.exp(negs_j), torch.exp(negs_k)

        # Compute InfoNCE
        denominator_final = pos_final_exp + negs_final_exp.sum(-1)
        denominator_j = pos_j_exp + negs_j_exp.sum(-1)
        denominator_k = pos_k_exp + negs_k_exp.sum(-1)

        nce_final = -torch.log(self.eps + (pos_final_exp / (self.eps + denominator_final)))
        nce_j = -torch.log(self.eps + (pos_j_exp / (self.eps + denominator_j)))
        nce_k = -torch.log(self.eps + (pos_k_exp / (self.eps + denominator_k)))

        # Compute "accuracy"
        acc_final = (1.0 * (negs_final_exp.max(dim=-1)[0] < pos_final_exp)).mean()
        acc_j = (1.0 * (negs_j_exp.max(dim=-1)[0] < pos_j_exp)).mean()
        acc_k = (1.0 * (negs_k_exp.max(dim=-1)[0] < pos_k_exp)).mean()
        acc = (acc_final + acc_j + acc_k) / 3
        nce = (nce_final + nce_j + nce_k) / 3

        return nce.mean(), acc

    def configure_optimizer(self) -> Tuple[torch.optim.Optimizer, Callable[[int, float], float]]:
        # Short-Circuit on Valid Optimizers
        if self.optimizer not in ["adam"]:
            raise NotImplementedError(f"Optimizer `{self.optimizer}` not supported - try [`adam`] instead!")

        # Create Optimizer and (No-Op) LR Scheduler
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        update_lr = get_lr_update(
            optimizer, schedule="none", lr=self.lr, min_lr=self.lr, warmup_epochs=-1, max_epochs=-1
        )
        return optimizer, update_lr
